from marshmallow.fields import Field

from archive_query_log.legacy.model import HighlightedText


class HighlightedTextField(Field):
    """
    Field that serializes to an HTML string and deserializes
    to a highlighted text.
    """

    def _serialize(self, value: HighlightedText, attr, obj, **kwargs):
        if value is None:
            return None
        return value.html

    def _deserialize(self, value: str, attr, data, **kwargs):
        if value is None:
            return None
        return HighlightedText(value)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_ebay_se_xing_mei_jia_wei_vxin_dun35358_1544323503():
    verify_serp_parsing(
        "https://web.archive.org/web/20181209024503id_/https://www.ebay.ca/sch/i.html?_nkw=%E8%89%B2%E6%80%A7%E5%A6%B9%E3%80%90%E5%8A%A0%E5%BE%AEv%E4%BF%A1%3Adun35358%E3%80%91%E2%8A%99",
        "ebay",
    )


def test_parse_query_ebay_shan_dong_11xuan_5_360cai_piao_wei_xin_weiwei776699_1538125942():
    verify_serp_parsing(
        "https://web.archive.org/web/20180928091222id_/https://www.ebay.ph/sch/i.html?_nkw=%E5%B1%B1%E4%B8%9C11%E9%80%895+360%E5%BD%A9%E7%A5%A8%2B%E3%80%96%E5%BE%AE%E4%BF%A1%3Aweiwei776699%E3%80%97%E2%94%A0",
        "ebay",
    )


def test_parse_query_ebay_smartphones_1334931269():
    verify_serp_parsing(
        "https://web.archive.org/web/20120420141429id_/http://www.ebay.de/sch/i.html?_nkw=Smartphones&_sacat=132286&rt=nc",
        "ebay",
    )


def test_parse_query_ebay_srch_str_1641320481():
    verify_serp_parsing(
        "https://web.archive.org/web/20220104182121id_/https://www.ebay.com.sg/sch/i.html?_nkw={srch_str}&ssPageName=GSTL",
        "ebay",
    )


def test_parse_query_ebay_liu_he_cai_gua_pai_wan_zheng_ban_wei_vxin_weiwei776699_1565563054():
    verify_serp_parsing(
        "https://web.archive.org/web/20190811223734id_/https://www.ebay.com.sg/sch/i.html?_nkw=%e5%85%ad%e5%90%88%e5%bd%a9%e6%8c%82%e7%89%8c%e5%ae%8c%e6%95%b4%e6%9d%bf%c2%b1%e3%80%90%e5%be%aev%e4%bf%a1weiwei776699%e3%80%91%e2%97%86",
        "ebay",
    )


def test_parse_query_ebay_1_4_1566151598():
    verify_serp_parsing(
        "https://web.archive.org/web/20190818180638id_/https://www.ebay.com/sch/i.html?rt=nc&_fsrp=1&_nkw=1+%3A+4+&_sacat=14964&Connector%2520A=RCA%252FPhono%2520Male",
        "ebay",
    )


def test_parse_query_ebay_hu_bei_kuai_san_zhong_jiang_zhu_shou_jia_wei_vxin_weiwei776699_1548361753():
    verify_serp_parsing(
        "https://web.archive.org/web/20190124202913id_/https://www.ebay.ph/sch/i.html?_nkw=%e6%b9%96%e5%8c%97%e5%bf%ab%e4%b8%89%e4%b8%ad%e5%a5%96%e5%8a%a9%e6%89%8b%e5%8a%a0%e3%80%8e%e5%be%aev%e4%bf%a1weiwei776699%e3%80%8f%e2%88%9a",
        "ebay",
    )


def test_parse_query_ebay_imaginext_batman_motorcycle_1597143554():
    verify_serp_parsing(
        "https://web.archive.org/web/20200811105914id_/https://www.ebay.com/sch/i.html?_nkw=imaginext+batman+motorcycle&rt=nc&_dcat=50305&_sacat=50305&vbn_id=7023320428&Featured%2520Refinements=Imaginext%2520Figure&_fsrp=1",
        "ebay",
    )


def test_parse_query_ebay_victoria_s_secret_1559541146():
    verify_serp_parsing(
        "https://web.archive.org/web/20190603055226id_/https://www.ebay.co.uk/sch/i.html?_nkw=Victoria%27s+Secret&rt=nc&_dcat=11514&_sacat=-1&cc_bn_id=21837960&_aspectname=aspect-Features&_fsrp=1",
        "ebay",
    )


def test_parse_query_ebay_liu_he_cai_2o2526_wei_xin_weiwei776699_1562672411():
    verify_serp_parsing(
        "https://web.archive.org/web/20190709114011id_/https://www.ebay.nl/sch/i.html?_nkw=%e5%85%ad%e5%90%88%e5%bd%a92o2526%2b%e3%80%8e%e5%be%ae%e4%bf%a1%3aweiwei776699%e3%80%8f%e2%96%b3",
        "ebay",
    )



from typing import Dict, Any, List
from urllib.parse import urlparse

from click import Parameter, Context
from click.shell_completion import CompletionItem
from click.types import StringParamType, Path, Choice


class UrlParam(StringParamType):
    name = "url"

    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        if value is None:
            return None
        tokens = urlparse(value)
        if not tokens.scheme or not tokens.netloc:
            self.fail(f"{value} is not a valid URL", param, ctx)
        return value


URL = UrlParam()

PathParam = Path


class ServiceChoice(Choice):

    def __init__(self) -> None:
        super().__init__(choices=[], case_sensitive=False)

    def _ensure_choices(self):
        if len(self.choices) == 0:
            from archive_query_log.legacy.config import SERVICES
            self.choices = sorted(SERVICES.keys())

    def to_info_dict(self) -> Dict[str, Any]:
        self._ensure_choices()
        return super().to_info_dict()

    def get_metavar(self, param: Parameter) -> str:
        self._ensure_choices()
        return super().get_metavar(param)

    def get_missing_message(self, param: Parameter) -> str:
        self._ensure_choices()
        return super().get_missing_message(param)

    def convert(
            self,
            value: Any,
            param: Parameter | None,
            ctx: Context | None,
    ) -> Any:
        self._ensure_choices()
        return super().convert(value, param, ctx)

    def __repr__(self) -> str:
        self._ensure_choices()
        return super().__repr__()

    def shell_complete(
            self,
            ctx: Context,
            param: Parameter,
            incomplete: str,
    ) -> List[CompletionItem]:
        self._ensure_choices()
        return super().shell_complete(ctx, param, incomplete)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_imgur_search_term_string_1565643838():
    verify_serp_parsing(
        "https://web.archive.org/web/20190812210358id_/https://imgur.com/search?q={search_term_string}",
        "imgur",
    )


def test_parse_query_imgur_search_term_string_1547858079():
    verify_serp_parsing(
        "https://web.archive.org/web/20190119003439id_/https://imgur.com/search?q={search_term_string}",
        "imgur",
    )



from dataclasses import dataclass, field
from functools import cached_property
from os import environ
from pathlib import Path
from typing import Iterable, Any

from elasticsearch import Elasticsearch
from elasticsearch.helpers import streaming_bulk
from pyrate_limiter import Limiter, RequestRate, Duration
from requests import Session
from requests_ratelimiter import LimiterAdapter
from urllib3 import Retry
from warc_cache import WarcCacheStore
from warc_s3 import WarcS3Store

from archive_query_log import __version__ as version


@dataclass(frozen=True)
class EsConfig:
    host: str = field(default_factory=lambda: environ["ELASTICSEARCH_HOST"])
    port: int = field(default_factory=lambda: int(environ["ELASTICSEARCH_PORT"]))
    username: str = field(default_factory=lambda: environ["ELASTICSEARCH_USERNAME"])
    password: str = field(default_factory=lambda: environ["ELASTICSEARCH_PASSWORD"])
    max_retries: int = 5
    bulk_chunk_size: int = 500
    bulk_max_chunk_bytes: int = 100 * 1024 * 1024
    bulk_initial_backoff: int = 2
    bulk_max_backoff: int = 60
    index_archives: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_ARCHIVES"]
    )
    index_providers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_PROVIDERS"]
    )
    index_sources: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_SOURCES"]
    )
    index_captures: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_CAPTURES"]
    )
    index_serps: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_SERPS"]
    )
    index_results: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_RESULTS"]
    )
    index_url_query_parsers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_URL_QUERY_PARSERS"]
    )
    index_url_page_parsers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_URL_PAGE_PARSERS"]
    )
    index_url_offset_parsers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_URL_OFFSET_PARSERS"]
    )
    index_warc_query_parsers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_WARC_QUERY_PARSERS"]
    )
    index_warc_snippets_parsers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_WARC_SNIPPETS_PARSERS"]
    )
    index_warc_direct_answers_parsers: str = field(
        default_factory=lambda: environ[
            "ELASTICSEARCH_INDEX_WARC_DIRECT_ANSWERS_PARSERS"
        ]
    )
    index_warc_main_content_parsers: str = field(
        default_factory=lambda: environ["ELASTICSEARCH_INDEX_WARC_MAIN_CONTENT_PARSERS"]
    )

    @cached_property
    def client(self) -> Elasticsearch:
        return Elasticsearch(
            hosts=f"https://{self.host}:{self.port}",
            http_auth=(self.username, self.password),
            timeout=60,
            max_retries=self.max_retries,
            retry_on_status=(502, 503, 504),
            retry_on_timeout=True,
        )

    def streaming_bulk(
        self,
        actions: Iterable[dict],
    ) -> Iterable[tuple[bool, Any]]:
        return streaming_bulk(
            client=self.client,
            actions=actions,
            chunk_size=self.bulk_chunk_size,
            max_chunk_bytes=self.bulk_max_chunk_bytes,
            initial_backoff=self.bulk_initial_backoff,
            max_backoff=self.bulk_max_backoff,
            max_retries=self.max_retries,
            raise_on_error=True,
            raise_on_exception=True,
            yield_ok=True,
        )

    def bulk(self, actions: Iterable[dict]) -> None:
        for _ in self.streaming_bulk(actions):
            pass


@dataclass(frozen=True)
class S3Config:
    endpoint_url: str = field(default_factory=lambda: environ["S3_ENDPOINT_URL"])
    access_key: str = field(default_factory=lambda: environ["S3_ACCESS_KEY"])
    secret_key: str = field(default_factory=lambda: environ["S3_SECRET_KEY"])
    bucket_name: str = field(default_factory=lambda: environ["S3_BUCKET_NAME"])

    @cached_property
    def warc_store(self) -> WarcS3Store:
        return WarcS3Store(
            endpoint_url=self.endpoint_url,
            access_key=self.access_key,
            secret_key=self.secret_key,
            bucket_name=self.bucket_name,
            max_file_size=1_000_000_000,
            quiet=False,
        )


@dataclass(frozen=True)
class HttpConfig:
    max_retries: int = 5

    @cached_property
    def session(self) -> Session:
        session = Session()
        session.headers.update(
            {
                "User-Agent": f"AQL/{version} (Webis group)",
            }
        )
        _retries = Retry(
            total=20,
            connect=5,
            read=5,
            redirect=10,
            status=10,
            backoff_factor=1,
            status_forcelist=[502, 503, 504],
            respect_retry_after_header=True,
        )
        _limiter = Limiter(
            RequestRate(1, Duration.SECOND * 10),
        )
        _adapter = LimiterAdapter(
            max_retries=_retries,
            limiter=_limiter,
            per_host=True,
        )
        # noinspection HttpUrlsUsage
        session.mount("http://", _adapter)
        session.mount("https://", _adapter)
        return session

    @cached_property
    def session_no_retry(self) -> Session:
        session = Session()
        session.headers.update(
            {
                "User-Agent": f"AQL/{version} (Webis group)",
            }
        )
        _limiter = Limiter(
            RequestRate(1, Duration.SECOND * 10),
        )
        _adapter = LimiterAdapter(
            limiter=_limiter,
        )
        # noinspection HttpUrlsUsage
        session.mount("http://", _adapter)
        session.mount("https://", _adapter)
        return session


@dataclass(frozen=True)
class WarcCacheConfig:
    path_serps: Path = field(
        default_factory=lambda: Path(environ["WARC_CACHE_PATH_SERPS"])
    )
    # path_results: Path = field(
    #     default_factory=lambda: Path(environ["WARC_CACHE_PATH_RESULTS"])
    # )

    @cached_property
    def store_serps(self) -> WarcCacheStore:
        return WarcCacheStore(
            cache_dir_path=self.path_serps,
            max_file_records=100,
            read_all_min_accumulated_bytes=1_000_000_000,
            read_all_include_temporary_files=False,
            quiet=False,
        )

    # @cached_property
    # def store_results(self) -> WarcCacheStore:
    #     return WarcCacheStore(
    #         cache_dir_path=self.path_results,
    #         max_file_records=100,
    #         read_all_min_accumulated_bytes=1_000_000_000,
    #         read_all_include_temporary_files=False,
    #         quiet=False,
    #     )


@dataclass(frozen=True)
class Config:
    es: EsConfig = field(default_factory=EsConfig)
    s3: S3Config = field(default_factory=S3Config)
    http: HttpConfig = field(default_factory=HttpConfig)
    warc_cache: WarcCacheConfig = field(default_factory=WarcCacheConfig)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_qq_danil_kozlovsky_1360453772():
    verify_serp_parsing(
        "https://web.archive.org/web/20130209234932id_/http://v.qq.com/search.html?pagetype=3&stag=word.tag&ms_key=danil%20kozlovsky",
        "qq",
    )


def test_parse_query_qq_feng_mi_lian_1380895166():
    verify_serp_parsing(
        "https://web.archive.org/web/20131004135926id_/http://v.qq.com/search.html?pagetype=3&stag=word.tag&ms_key=%E5%B3%B0%E5%B9%82%E6%81%8B",
        "qq",
    )


def test_parse_query_qq_ji_lin_yan_ji_1360789033():
    verify_serp_parsing(
        "https://web.archive.org/web/20130213205713id_/http://v.qq.com:80/search.html?pagetype=3&ms_key=%E5%90%89%E6%9E%97%E5%BB%B6%E5%90%89",
        "qq",
    )


def test_parse_query_qq_zhong_nian_wei_ji_1408289827():
    verify_serp_parsing(
        "https://web.archive.org/web/20140817153707id_/http://v.qq.com/search.html?pagetype=3&stag=word.tag&ms_key=%E4%B8%AD%E5%B9%B4%E5%8D%B1%E6%9C%BA",
        "qq",
    )


def test_parse_query_qq_jie_ke_luo_de_wei_er_1445892389():
    verify_serp_parsing(
        "https://web.archive.org/web/20151026204629id_/http://v.qq.com/search.html?pagetype=3&stag=txt.playpage.sports&ms_key=%E6%9D%B0%E5%85%8B.%E7%BD%97%E5%BE%B7%E7%BB%B4%E5%B0%94",
        "qq",
    )


def test_parse_query_qq_lin_xi_tong_1507483207():
    verify_serp_parsing(
        "https://web.archive.org/web/20171008172007id_/http://v.qq.com/search.html?pagetype=3&stj2=tiket&ms_key=%E6%9E%97%E7%86%B9%E7%9E%B3",
        "qq",
    )


def test_parse_query_qq_shen_shan_zheng_er_lang_1405119212():
    verify_serp_parsing(
        "https://web.archive.org/web/20140711225332id_/http://v.qq.com//search.html?pagetype=3&ms_key=%E7%A5%9E%E5%B1%B1%E5%BE%81%E4%BA%8C%E9%83%8E",
        "qq",
    )


def test_parse_query_qq_james_boshier_1449178049():
    verify_serp_parsing(
        "https://web.archive.org/web/20151203212729id_/http://v.qq.com//search.html?pagetype=3&ms_key=James+Boshier",
        "qq",
    )


def test_parse_query_qq_ji_zhou_dao_ttglao_hu_ji_ping_tai_guan_wang_x_fa33n_comx_1488991663():
    verify_serp_parsing(
        "https://web.archive.org/web/20170308164743id_/http://v.qq.com/search.html?pagetype=3&stj2=search.search&stag=txt.index&ms_key=%E6%B5%8E%E5%B7%9E%E5%B2%9Bttg%E8%80%81%E8%99%8E%E6%9C%BA%E5%B9%B3%E5%8F%B0%E5%AE%98%E7%BD%91%E3%80%93FA33N.COM%E3%80%93",
        "qq",
    )


def test_parse_query_qq_hua_ze_xiang_cai_1408309927():
    verify_serp_parsing(
        "https://web.archive.org/web/20140817211207id_/http://v.qq.com//search.html?pagetype=3&ms_key=%E8%8A%B1%E6%B3%BD%E9%A6%99%E8%8F%9C",
        "qq",
    )


def test_parse_query_qq_zhang_dong_jian_1446133153():
    verify_serp_parsing(
        "https://web.archive.org/web/20151029153913id_/http://v.qq.com//search.html?pagetype=3&ms_key=%E5%BC%A0%E4%B8%9C%E5%81%A5",
        "qq",
    )


def test_parse_query_qq_jin_xiu_long_1320298924():
    verify_serp_parsing(
        "https://web.archive.org/web/20111103054204id_/http://v.qq.com/search.html?pagetype=3&ms_key=%E9%87%91%E7%A7%80%E9%BE%99",
        "qq",
    )


def test_parse_query_qq_xing_xing_di_qiu_2_1582812539():
    verify_serp_parsing(
        "https://web.archive.org/web/20200227140859id_/http://v.qq.com:80/search.html?ms_key=%E8%A1%8C%E6%98%9F%E5%9C%B0%E7%90%832",
        "qq",
    )


def test_parse_query_qq_turn_that_finger_around_1324266860():
    verify_serp_parsing(
        "https://web.archive.org/web/20111219035420id_/http://v.qq.com:80/search.html?pagetype=3&ms_key=Turn+That+Finger+Around",
        "qq",
    )


def test_parse_query_qq_statue_1536122094():
    verify_serp_parsing(
        "https://web.archive.org/web/20180905043454id_/http://v.qq.com:80/search.html?pagetype=3&stj2=search.search&stag=txt.index&ms_key=statue",
        "qq",
    )


def test_parse_query_qq_erin_1359911760():
    verify_serp_parsing(
        "https://web.archive.org/web/20130203171600id_/http://v.qq.com//search.html?pagetype=3&ms_key=Erin",
        "qq",
    )


def test_parse_query_qq_jin_cheng_wu_2_1319745059():
    verify_serp_parsing(
        "https://web.archive.org/web/20111027195059id_/http://v.qq.com/search.html?pagetype=3&ms_key=%E9%87%91%E5%9F%8E%E6%AD%A6",
        "qq",
    )


def test_parse_query_qq_tag_encode_1341871817():
    verify_serp_parsing(
        "https://web.archive.org/web/20120709221017id_/http://v.qq.com/search.html?pagetype=3&ms_key=${tag|encode}",
        "qq",
    )


def test_parse_query_qq_niamh_cusack_1319741306():
    verify_serp_parsing(
        "https://web.archive.org/web/20111027213506id_/http://v.qq.com/search.html?pagetype=3&ms_key=Niamh+Cusack",
        "qq",
    )



from click import group, option

from archive_query_log.cli.util import pass_config
from archive_query_log.config import Config
from archive_query_log.orm import Serp, Result


@group()
def serps():
    pass


@serps.group()
def parse():
    pass


@parse.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Parse SERP URLs for only a limited number of captures, and prefetch that batch to parse.",
)
def url_query(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.parsers.url_query import parse_serps_url_query

    Serp.init(
        using=config.es.client,
        index=config.es.index_serps,
    )
    parse_serps_url_query(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@parse.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Parse only a limited number of SERPs, and prefetch that batch to parse.",
)
def url_page(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.parsers.url_page import parse_serps_url_page

    parse_serps_url_page(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@parse.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Parse only a limited number of SERPs, and prefetch that batch to parse.",
)
def url_offset(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.parsers.url_offset import parse_serps_url_offset

    parse_serps_url_offset(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@parse.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Parse only a limited number of SERPs, and prefetch that batch to parse.",
)
def warc_query(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.parsers.warc_query import parse_serps_warc_query

    parse_serps_warc_query(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@parse.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Parse only a limited number of SERPs, and prefetch that batch to parse.",
)
def warc_snippets(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.parsers.warc_snippets import parse_serps_warc_snippets

    Result.init(
        using=config.es.client,
        index=config.es.index_results,
    )
    parse_serps_warc_snippets(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@parse.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Parse only a limited number of SERPs, and prefetch that batch to parse.",
)
def warc_direct_answers(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.parsers.warc_direct_answers import (
        parse_serps_warc_direct_answers,
    )

    parse_serps_warc_direct_answers(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@serps.group()
def download():
    pass


@download.command(
    name="warc", help="Download archived documents of captures as WARC to a file cache."
)
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Download only a limited number of SERPs, and prefetch that batch to parse.",
)
def download_warc(config: Config, prefetch_limit: int | None = None) -> None:
    from archive_query_log.downloaders.warc import download_serps_warc

    download_serps_warc(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@serps.group()
def upload():
    pass


@upload.command(
    name="warc",
    help="Upload WARCs of archived documents of captures to S3 and update the index.",
)
@pass_config
def upload_warc(config: Config) -> None:
    from archive_query_log.downloaders.warc import upload_serps_warc

    upload_serps_warc(config)



from io import IOBase, TextIOWrapper
from typing import IO, Iterator

_LINE_COUNT_BUFFER_SIZE = 1024 * 1024


def _chunks(reader: IO[bytes] | IOBase) -> Iterator[bytes]:
    buffer = reader.read(_LINE_COUNT_BUFFER_SIZE)
    while buffer:
        yield buffer
        buffer = reader.read(_LINE_COUNT_BUFFER_SIZE)


def count_lines(file: IO[bytes] | IOBase) -> int:
    return sum(buffer.count(b"\n") for buffer in _chunks(file))


def text_io_wrapper(file: IO[bytes] | IOBase) -> IO[str]:
    return TextIOWrapper(file)  # type: ignore



from dataclasses import dataclass
from io import BytesIO
from itertools import count, groupby, chain
from pathlib import Path
from random import Random
from tempfile import TemporaryFile
from typing import Sequence, NamedTuple, Iterable
from urllib.parse import quote, parse_qsl

from aiohttp import ClientResponseError
from aiohttp_retry import RetryClient
from asyncio_pool import AioPool
from tqdm.auto import tqdm
from warcio import WARCWriter, StatusAndHeaders

from archive_query_log.legacy.model import ArchivedUrl, Service, \
    ArchivedQueryUrl
from archive_query_log.legacy.queries.iterable import ArchivedQueryUrls
from archive_query_log.legacy.serps.iterable import ArchivedParsedSerps
from archive_query_log.legacy.util.archive_http import archive_http_client


class _CdxPage(NamedTuple):
    input_path: Path
    output_path: Path


class _CdxUrl(NamedTuple):
    archived_url: ArchivedUrl
    output_path: Path


@dataclass(frozen=True)
class WebArchiveWarcDownloader:
    """
    Download WARC files for archived URLs from the Web Archive.

    The downloader will retry requests with a backoff and will continue
    downloading URLs even if some URLs fail.
    """

    max_file_size: int = 1_000_000_000  # 1GB
    """
    Maximum number of bytes to write to a single WARC file.
    """
    verbose: bool = False

    @staticmethod
    def _check_download_path(download_path: Path):
        download_path.mkdir(parents=True, exist_ok=True)
        if not download_path.is_dir():
            raise ValueError(
                f"Download path must be a directory: {download_path}"
            )

    @staticmethod
    def _lock_path(download_path: Path) -> Path:
        path = download_path / ".lock"
        path.touch(exist_ok=True)
        return path

    def _is_url_downloaded(self, url: _CdxUrl) -> bool:
        if not url.output_path.exists():
            return False
        archive_url = url.archived_url.raw_archive_url
        with self._lock_path(url.output_path).open("rt") as file:
            return any(
                line.strip() == archive_url
                for line in file
            )

    def _set_url_downloaded(self, url: _CdxUrl):
        archive_url = url.archived_url.raw_archive_url
        with self._lock_path(url.output_path).open("at") as file:
            file.write(f"{archive_url}\n")

    def _next_available_file_path(
            self,
            download_path: Path,
            buffer_size: int,
    ) -> Path:
        WebArchiveWarcDownloader._check_download_path(download_path)
        for index in count():
            name = f"{index:010}.warc.gz"
            path = download_path / name
            if not path.exists():
                path.touch()
                return path
            else:
                file_size = path.stat().st_size
                if file_size + buffer_size <= self.max_file_size:
                    return path
        raise RuntimeError("All available file paths are filled.")

    async def _download(
            self,
            urls: Iterable[_CdxUrl],
    ) -> None:
        """
        Download WARC files for archived URLs from the Web Archive.
        """

        urls = [
            url
            for url in urls
            if not self._is_url_downloaded(url)
        ]
        if len(urls) == 0:
            return

        progress = None
        if self.verbose:
            progress = tqdm(
                total=len(urls),
                desc="Download archived URLs",
                unit="URL",
            )

        async with archive_http_client(limit=100) as client:
            pool = AioPool(size=100)  # avoid creating too many tasks at once

            async def download_single(url: _CdxUrl):
                return await self._download_single(client, url, progress)

            await pool.map(download_single, urls)

    async def download(
            self,
            download_path: Path,
            archived_urls: Iterable[ArchivedUrl],
    ) -> None:
        """
        Download WARC files for archived URLs from the Web Archive.
        """
        await self._download([
            _CdxUrl(url, download_path)
            for url in archived_urls
        ])

    async def _download_single(
            self,
            client: RetryClient,
            url: _CdxUrl,
            progress: tqdm | None = None,
    ) -> bool:
        if self._is_url_downloaded(url):
            if progress is not None:
                progress.update()
            return True
        archive_url = url.archived_url.raw_archive_url
        url_headers = {
            "Archived-URL": url.archived_url.schema().dumps(url.archived_url),
        }
        try:
            async with client.get(archive_url) as response:
                response.raise_for_status()
                with TemporaryFile() as tmp_file:
                    writer = WARCWriter(tmp_file, gzip=True)
                    # noinspection PyProtectedMember
                    version = client._client.version
                    protocol = f"HTTP/{version[0]}.{version[1]}"
                    request_record = writer.create_warc_record(
                        uri=str(response.request_info.url),
                        record_type="request",
                        http_headers=StatusAndHeaders(
                            statusline=" ".join((
                                response.request_info.method,
                                response.request_info.url.path,
                                protocol,
                            )),
                            headers=response.request_info.headers,
                            protocol=protocol,
                        ),
                        warc_headers_dict={**url_headers},
                    )
                    writer.write_record(request_record)

                    protocol = f"HTTP/{response.version}"
                    reason = str(response.reason)
                    response_record = writer.create_warc_record(
                        uri=str(response.url),
                        record_type="response",
                        http_headers=StatusAndHeaders(
                            statusline=" ".join((
                                protocol,
                                str(response.status),
                                reason,
                            )),
                            headers=response.headers,
                            protocol=protocol
                        ),
                        payload=BytesIO(await response.content.read()),
                        length=response.content_length,
                        warc_headers_dict={**url_headers},
                    )
                    writer.write_record(response_record)
                    tmp_file.flush()
                    tmp_size = tmp_file.tell()
                    tmp_file.seek(0)
                    file_path = self._next_available_file_path(
                        url.output_path,
                        tmp_size,
                    )
                    with file_path.open("ab") as file:
                        tmp = tmp_file.read()
                        if not len(tmp) == tmp_size:
                            raise RuntimeError("Invalid buffer size.")
                        file.write(tmp)
                    self._set_url_downloaded(url)
                    return True
        except ClientResponseError:
            return False
        except BaseException as e:
            raise e
        finally:
            if progress is not None:
                progress.update()

    def _service_pages(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None,
            cdx_page: int | None,
            snippets: bool = False,
    ) -> Sequence[_CdxPage]:
        """
        List all items that need to be downloaded.
        """
        if snippets:
            input_format_path = data_directory / "archived-parsed-serps"
            output_format_path = data_directory / "archived-raw-search-results"
        else:
            input_format_path = data_directory / "archived-query-urls"
            output_format_path = data_directory / "archived-raw-serps"

        service_path = input_format_path / service.name

        if domain is not None:
            domain_paths = [service_path / domain]
        else:
            domain_paths = [
                path
                for path in service_path.iterdir()
                if path.is_dir()
            ]
            if focused:
                domain_paths = [
                    path
                    for path in domain_paths
                    if any(
                        path.name.endswith(quote(prefix, safe=""))
                        for prefix in service.focused_url_prefixes
                    )
                ]

        if cdx_page is not None:
            if domain is None:
                raise RuntimeError(
                    "Domain must be specified when page is specified.")
            if len(domain_paths) < 1:
                raise RuntimeError(
                    "There must be exactly one domain path.")
            cdx_page_paths = [domain_paths[0] / f"{cdx_page:010}.jsonl.gz"]
        else:
            cdx_page_paths = [
                path
                for domain_path in domain_paths
                for path in domain_path.iterdir()
                if (
                        path.is_file() and
                        len(path.name.removesuffix(".jsonl.gz")) == 10 and
                        path.name.removesuffix(".jsonl.gz").isdigit()
                )
            ]

        pages = (
            _CdxPage(
                input_path=cdx_page_path,
                output_path=output_format_path / cdx_page_path.relative_to(
                    input_format_path
                ).with_name(cdx_page_path.name.removesuffix(".jsonl.gz")),
            )
            for cdx_page_path in cdx_page_paths
        )
        return [page for page in pages if page.input_path.exists()]

    @staticmethod
    def _canonical_url(urls: Iterable[_CdxUrl]) -> _CdxUrl:
        """
        First URL, sorted by URL query string length, then by URL length,
        then by URL.
        """
        urls = sorted(
            urls,
            key=lambda url: url.archived_url.url
        )
        urls = sorted(
            urls,
            key=lambda url: len(url.archived_url.url)
        )
        urls = sorted(
            urls,
            key=lambda url: len(parse_qsl(url.archived_url.split_url.query))
        )
        return urls[0]

    @staticmethod
    def _deduplicate_urls(
            urls: Iterable[_CdxUrl],
            snippets: bool,
    ) -> list[_CdxUrl]:
        if snippets:
            return list(urls)
        if not all(
            isinstance(url.archived_url, ArchivedQueryUrl)
            for url in urls
        ):
            return list(urls)
        urls = sorted(
            urls,
            key=lambda url: (
                url.archived_url.query
                if isinstance(url.archived_url, ArchivedQueryUrl) else "")
        )
        grouped_query_urls = groupby(
            urls,
            key=lambda url: (
                url.archived_url.query
                if isinstance(url.archived_url, ArchivedQueryUrl) else "")
        )
        return [
            WebArchiveWarcDownloader._canonical_url(urls)
            for query, urls in grouped_query_urls
        ]

    @staticmethod
    def _page_urls(
            page: _CdxPage,
            focused: bool,
            snippets: bool,
    ) -> Iterable[_CdxUrl]:
        urls: Iterable[_CdxUrl]
        if snippets:
            urls = (
                _CdxUrl(url, page.output_path)
                for serp in ArchivedParsedSerps(page.input_path)
                for url in serp.results
            )
        else:
            urls = (
                _CdxUrl(url, page.output_path)
                for url in ArchivedQueryUrls(page.input_path)
            )
        if focused:
            urls = WebArchiveWarcDownloader._deduplicate_urls(urls, snippets)
        return urls

    async def download_service(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None = None,
            cdx_page: int | None = None,
            snippets: bool = False,
    ):
        pages_list: Sequence[_CdxPage] = self._service_pages(
            data_directory=data_directory,
            focused=focused,
            service=service,
            domain=domain,
            cdx_page=cdx_page,
            snippets=snippets,
        )

        if len(pages_list) == 0:
            return

        pages: Iterable[_CdxPage] = pages_list
        if focused:
            # noinspection PyTypeChecker
            pages = tqdm(
                pages,
                desc="Deduplicate query URLs",
                unit="page",
            )

        cdx_urls: Sequence[_CdxUrl] = list(chain.from_iterable(
            self._page_urls(page, focused, snippets)
            for page in pages
        ))

        if focused:
            archived_urls_list = self._deduplicate_urls(
                cdx_urls, snippets)
            sample_size = min(len(cdx_urls), 75_000)
            random = Random(0)  # nosec: B311
            cdx_urls = random.sample(archived_urls_list, sample_size)

        await self._download(cdx_urls)



import re
import pandas as pd
from tqdm import tqdm
from requests_html import HTMLSession
from bs4 import BeautifulSoup
import argparse

pattern = re.compile(r'(?i).*search.*')


# noinspection PyUnresolvedReferences,PyBroadException
class SearchFormIdentifier:
    """
    Class that takes in a CSV-file containing services. The file needs
    to have the following columns:
    - rank
    - service (name with TLD e.g. 'amazon.com')
    - TLD
    The process_services method will look for input fields, search forms
    and search divs on the corresponding website,
    indicate if the service has such a field as well as document
    any HMTL-snippet found during the search
    """

    def __init__(self, csv_file='./ranked_services.csv', outfile_num=0,
                 start_row=0, end_row=None):
        self.out_df = None
        self.df = pd.read_csv(csv_file, header=None)
        self.df.columns = ['rank', 'service', 'tld']

        if end_row is None:
            end_row = len(self.df)
        self.df = self.df[start_row:end_row]

        self.out_file = f'./search_forms_{outfile_num}.csv'
        self.session = HTMLSession()

    def process_services(self):
        """
        Use pd.DataFrame.progress_apply to run the method check_url
        on all services.
        Store the results in an attribute out_df and save it in a csv-File.
        """
        tqdm.pandas()
        self.out_df = self.df.copy(deep=True)

        # Search for the HTML elements and split the results
        # into separate columns
        self.out_df['tmp'] = self.df['service'].progress_apply(
            func=self.check_url)
        self.out_df[['input', 'search_form', 'search_div', 'input_snippets',
                     'form_snippets', 'div_snippets']] = \
            pd.DataFrame(self.out_df['tmp'].tolist(), index=self.out_df.index)
        self.out_df.drop('tmp', axis=1, inplace=True)
        self.out_df.to_csv(self.out_file)

    def check_url(self, url: str):
        """
        Method to take in a service URL and look for
        any relevant tags in the HTML.
        It will look for any input tag and div or form with an id that
        has the term 'search' in it

        :param      url:                Service URL to be processed
        :return:    found_input:        Boolean
                    found_search_form:  Boolean
                    found_search_div:   Boolean
                    input_snippets:     List of identified HTML-snippets
                    form_snippets:      List of identified HTML-snippets
                    div_snippets:       List of identified HTML-snippets
        """

        url = 'https://' + url if 'http' not in url else url

        # If regular requesting fails, try to get a recent snapshot
        # in the internet archive
        try:
            response = self.session.get(url, timeout=10)
        except Exception:
            try:
                response = self.get_internet_archive_html(url=url)
            except Exception:
                return None, None, None

        try:
            html = response.html.html
        except Exception:
            return None, None, None

        # Render JavaScript necessary
        if '<script>' in html:
            try:
                response.html.render(timeout=10)
                html = response.html.html
            except Exception:
                return None, None, None

        # Look for elements with the pattern in them and save the snippets
        soup = BeautifulSoup(html, 'html.parser')
        found_input, input_snippets = find_input_tag(soup=soup)
        found_search_form, form_snippets = find_search_tag(
            soup=soup, tag='form')
        found_search_div, div_snippets = find_search_tag(soup=soup, tag='div')

        return found_input, found_search_form, found_search_div, \
            input_snippets, form_snippets, div_snippets

    def get_internet_archive_html(self, url: str, year=2022, byte_digits=4):
        """
        Method to get the response for a URL from the Internet Archive.
        It will return the most recent snapshot from the specified year
        with a minimum size

        :param url:             URL for which to find a snapshot
        :param year:            The desired year for snapshots
        :param byte_digits:     Defines the minimum size with a leading 5.
                                Examples byte_digits=4 -> Bytes >= 5,000;
                                byte_digits=5 -> Bytes >= 50,000
        :return:                The response from the Internet Archive Snapshot
        """
        # Get the snapshots with a minimum number of bytes
        # from the specified year and extract the timestamp of the most recent
        search_url = f'https://web.archive.org/cdx/search/cdx' \
                     f'?url={url}&fl=original,timestamp,length&from={year}' \
                     f'&filter=mimetype:text/html&filter=statuscode:200' \
                     f'&filter=length:.*[5-9][0-9]%7B{byte_digits - 1},%7D.*'
        try:
            response = self.session.get(search_url, timeout=10)
            snapshot_list = response.html.html.split('\n')
            timestamp = snapshot_list[-2].split(" ")[1]

            # Request the corresponding HTML
            ia_url = f'https://web.archive.org/web/{timestamp}/{url}/'
            return self.session.get(ia_url, timeout=10)
        except Exception as e:
            raise RuntimeError(
                'Failed to request an internet archive snapshot') from e

    def services_no_search(self):
        return self.out_df[(self.out_df['input'] is False) & (
                self.out_df['search_form'] is False) &
                           (self.out_df['search_div'] is False)]


def find_input_tag(soup: BeautifulSoup):
    """
    Function to look for HTML-Elements with the input tag
    :param      soup:           A bs4.BeautifulSoup instance with the HTML
        to conduct tag search on
    :return:    found:          Boolean to indicate if search was successful
                snippet_list:   List of matching HTML snippets that were found

    """
    snippet_list = soup.findAll(re.compile(r'(?i).*input.*'))
    found = True if len(snippet_list) > 0 else False
    return found, snippet_list


def find_search_tag(soup: BeautifulSoup, tag='form'):
    """
    Function to look for HTML-Elements with the specified tag
    that contain the pattern "search" in some way
    :param      soup:           A bs4.BeautifulSoup instance with the HTML
                                to conduct tag search on
                tag:            Name of relevant tags
    :return:    found:          Boolean to indicate if search was successful
                snippet_list:   List of matching HTML snippets that were found

    """
    snippet_list = soup.findAll(tag, {"id": pattern})
    found = True if len(snippet_list) > 0 else False
    return found, snippet_list


def main():
    # Parse input
    parser = argparse.ArgumentParser(
        prog='Search form identification',
        description='Takes in a CSV-File of services '
                    'and looks for search forms in their HTML')
    parser.add_argument('-f', '--csv_file', type=str)
    parser.add_argument('-o', '--outfile_num', type=str)
    parser.add_argument('-s', '--start_row', type=int)
    parser.add_argument('-e', '--end_row', type=int)
    args = parser.parse_args()

    # Set/Update default values
    csv_file = './alexa-top-1m-fused-domains-rrf-top-10000.csv' \
        if args.csv_file is None else args.csv_file
    outfile_num = "0" if args.outfile_num is None else args.outfile_num
    start_row = 0 if args.start_row is None else args.start_row
    end_row = None if args.end_row is None else args.end_row

    # Run the search for specified services
    identifier = SearchFormIdentifier(
        csv_file=csv_file, outfile_num=outfile_num,
        start_row=start_row, end_row=end_row
    )
    identifier.process_services()


if __name__ == "__main__":
    main()



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_yahoo_diver_lg_u8180_1620023310():
    verify_serp_parsing(
        "https://web.archive.org/web/20210503062830id_/https://it.search.yahoo.com/search?p=+diver+LG+U8180&fr=yfp-t-501&ei=UTF-8&rd=r1",
        "yahoo",
    )


def test_parse_query_yahoo_civil_code_1658358754():
    verify_serp_parsing(
        "https://web.archive.org/web/20220720231234id_/https://search.yahoo.com/search?p=Civil%20Code",
        "yahoo",
    )


def test_parse_query_yahoo_world_greek_kata_periokhe_boreia_amerike_kanadas_1647085990():
    verify_serp_parsing(
        "https://web.archive.org/web/20220312115310id_/https://search.yahoo.com/search?p=World+Greek+%CE%9A%CE%B1%CF%84%CE%AC+%CE%A0%CE%B5%CF%81%CE%B9%CE%BF%CF%87%CE%AE+%CE%92%CF%8C%CF%81%CE%B5%CE%B9%CE%B1+%CE%91%CE%BC%CE%B5%CF%81%CE%B9%CE%BA%CE%AE+%CE%9A%CE%B1%CE%BD%CE%B1%CE%B4%CE%AC%CF%82&ei=UTF-8&nojs=1",
        "yahoo",
    )


def test_parse_query_yahoo_lm_1376048795():
    verify_serp_parsing(
        "https://web.archive.org/web/20130809114635id_/http://search.yahoo.com/search?p=%3Blm",
        "yahoo",
    )


def test_parse_query_yahoo_ashland_oregon_1015423512():
    verify_serp_parsing(
        "https://web.archive.org/web/20020306140512id_/http://search.yahoo.com:80/search?p=ashland+oregon+",
        "yahoo",
    )


def test_parse_query_yahoo_lawyer_career_information_970996615():
    verify_serp_parsing(
        "https://web.archive.org/web/20001008091655id_/http://search.yahoo.com:80/search?p=lawyer+career+information",
        "yahoo",
    )


def test_parse_query_yahoo_world_hebrew_hbrh_shlvm_1489427128():
    verify_serp_parsing(
        "https://web.archive.org/web/20170313174528id_/https://search.yahoo.com/search?p=World+Hebrew+%D7%97%D7%91%D7%A8%D7%94+%D7%A9%D7%9C%D7%95%D7%9D",
        "yahoo",
    )


def test_parse_query_yahoo_john_mccain_956454509():
    verify_serp_parsing(
        "https://web.archive.org/web/20000423014829id_/http://search.yahoo.com:80/search?p=John+McCain&hn=48",
        "yahoo",
    )


def test_parse_query_yahoo_bc_gov_971492707():
    verify_serp_parsing(
        "https://web.archive.org/web/20001014030507id_/http://search.yahoo.com:80/search?p=bc+gov",
        "yahoo",
    )


def test_parse_query_yahoo_region_de_coquimbo_970960857():
    verify_serp_parsing(
        "https://web.archive.org/web/20001007232057id_/http://search.yahoo.com:80/search?p=region+de+coquimbo",
        "yahoo",
    )


def test_parse_query_yahoo_learn_to_focus_1619373855():
    verify_serp_parsing(
        "https://web.archive.org/web/20210425180415id_/https://search.yahoo.com/search?p=learn+to+focus&amp;fr2=sb-top&amp;sao=1",
        "yahoo",
    )


def test_parse_query_yahoo_castle_heights_angeles_1611631735():
    verify_serp_parsing(
        "https://web.archive.org/web/20210126032855id_/http://search.yahoo.com/search?fr=sbcfp&p=%22castle+heights%22+angeles",
        "yahoo",
    )


def test_parse_query_yahoo_dunwells_army_of_friends_1617218301():
    verify_serp_parsing(
        "https://web.archive.org/web/20210331191821id_/https://search.yahoo.com/search?p=dunwells+army+of+friends&fr=iphone&.tsrc=apple&pcarrier=AT&T&pmcc=310&pmnc=410",
        "yahoo",
    )


def test_parse_query_yahoo_sam_bush_1016625678():
    verify_serp_parsing(
        "https://web.archive.org/web/20020320120118id_/http://search.yahoo.com:80/search?p=Sam+Bush",
        "yahoo",
    )


def test_parse_query_yahoo_telekwiaciarnia_pl_1436841282():
    verify_serp_parsing(
        "https://web.archive.org/web/20150714023442id_/http://nl.search.yahoo.com/search?fr=yhs-invalid&p=telekwiaciarnia.pl",
        "yahoo",
    )


def test_parse_query_yahoo_jello_gelatin_971145484():
    verify_serp_parsing(
        "https://web.archive.org/web/20001010023804id_/http://search.yahoo.com:80/search?p=jello+gelatin",
        "yahoo",
    )


def test_parse_query_yahoo_molly_shannon_1640079691():
    verify_serp_parsing(
        "https://web.archive.org/web/20211221094131id_/https://search.yahoo.com/search?p=Molly%20Shannon",
        "yahoo",
    )


def test_parse_query_yahoo_dating_sites_1647656764():
    verify_serp_parsing(
        "https://web.archive.org/web/20220319022604id_/https://search.yahoo.com/search?ei=UTF-8&p=dating+sites&fr2=p%3As%2Cv%3Asfp%2Cm%3Aovncp&fr=sfp",
        "yahoo",
    )


def test_parse_query_yahoo_metallurgi_1042532635():
    verify_serp_parsing(
        "https://web.archive.org/web/20030114082355id_/http://search.yahoo.com:80/search?p=metallurgi",
        "yahoo",
    )


def test_parse_query_yahoo_futurama_1314932107():
    verify_serp_parsing(
        "https://web.archive.org/web/20110902025507id_/http://tv.yahoo.com:80/search?cs=bz&p=Futurama&fr=tv-tts",
        "yahoo",
    )


def test_parse_query_yahoo_gabel_970409689():
    verify_serp_parsing(
        "https://web.archive.org/web/20001001141449id_/http://search.yahoo.com:80/search?p=gabel",
        "yahoo",
    )


def test_parse_query_yahoo_gina_mckee_998889511():
    verify_serp_parsing(
        "https://web.archive.org/web/20010827051831id_/http://search.yahoo.com:80/search?p=Gina+McKee",
        "yahoo",
    )


def test_parse_query_yahoo_fruitarianism_970289154():
    verify_serp_parsing(
        "https://web.archive.org/web/20000930044554id_/http://search.yahoo.com:80/search?p=fruitarianism",
        "yahoo",
    )


def test_parse_query_yahoo_gamefaq_970449707():
    verify_serp_parsing(
        "https://web.archive.org/web/20001002012147id_/http://search.yahoo.com:80/search?p=gamefaq",
        "yahoo",
    )


def test_parse_query_yahoo_glass_tools_961515348():
    verify_serp_parsing(
        "https://web.archive.org/web/20000620153548id_/http://search.yahoo.com:80/search?p=glass+tools",
        "yahoo",
    )


def test_parse_query_yahoo_bill_callahan_american_football_1636808341():
    verify_serp_parsing(
        "https://web.archive.org/web/20211113125901id_/http://au.search.yahoo.com/search?p=Bill+Callahan+(American+football)&ei=UTF-8&d=%7B%22dn%22%3A%22yk%22%2C%22ykid%22%3A%22d9b7247f-1d93-4148-a7cc-d7447903c37a%22+%7D&fr2=p%3As%2Cv%3Aw%2Cm%3Akg-pasf%2Cct%3AGetPersonInfo&fr=fp-tts",
        "yahoo",
    )


def test_parse_query_yahoo_general_s_daughter_the_970749419():
    verify_serp_parsing(
        "https://web.archive.org/web/20001005123659id_/http://search.yahoo.com:80/search?p=General's+Daughter,+The",
        "yahoo",
    )


def test_parse_query_yahoo_golson_1017320115():
    verify_serp_parsing(
        "https://web.archive.org/web/20020328125515id_/http://search.yahoo.com:80/search?p=golson",
        "yahoo",
    )


def test_parse_query_yahoo_golden_books_971368161():
    verify_serp_parsing(
        "https://web.archive.org/web/20001012162921id_/http://search.yahoo.com:80/search?p=golden+books",
        "yahoo",
    )


def test_parse_query_yahoo_girlsandanimals_1039016411():
    verify_serp_parsing(
        "https://web.archive.org/web/20021204154011id_/http://search.yahoo.com:80/search?p=girlsandanimals",
        "yahoo",
    )


def test_parse_query_yahoo_ganze_filme_online_schauen_kostenlos_und_in_voller_lange_1376615487():
    verify_serp_parsing(
        "https://web.archive.org/web/20130816011127id_/http://search.yahoo.com/search?p=Ganze+Filme+online+schauen+kostenlos+und+in+voller+L%C3%A4nge",
        "yahoo",
    )


def test_parse_query_yahoo_full_speed_1011203807():
    verify_serp_parsing(
        "https://web.archive.org/web/20020116175647id_/http://search.yahoo.com:80/search?p=Full+Speed",
        "yahoo",
    )


def test_parse_query_yahoo_future_of_sport_971016077():
    verify_serp_parsing(
        "https://web.archive.org/web/20001008144117id_/http://search.yahoo.com:80/search?p=future+of+sport",
        "yahoo",
    )


def test_parse_query_yahoo_george_foreman_boxer_1040476713():
    verify_serp_parsing(
        "https://web.archive.org/web/20021221131833id_/http://search.yahoo.com:80/search?p=george+foreman+boxer",
        "yahoo",
    )


def test_parse_query_yahoo_gogi_grant_1041483222():
    verify_serp_parsing(
        "https://web.archive.org/web/20030102045342id_/http://search.yahoo.com:80/search?p=Gogi+Grant",
        "yahoo",
    )


def test_parse_query_yahoo_ghulam_sarwar_khan_1626337646():
    verify_serp_parsing(
        "https://web.archive.org/web/20210715082726id_/https://search.yahoo.com/search?p=Ghulam%20Sarwar%20Khan",
        "yahoo",
    )


def test_parse_query_yahoo_free_mail_centers_970285370():
    verify_serp_parsing(
        "https://web.archive.org/web/20000930034250id_/http://search.yahoo.com:80/search?p=FREE+MAIL+CENTERS",
        "yahoo",
    )


def test_parse_query_yahoo_games_970945905():
    verify_serp_parsing(
        "https://web.archive.org/web/20001007191145id_/http://search.yahoo.com:80/search?p=gAmes",
        "yahoo",
    )


def test_parse_query_yahoo_gamespot_portugal_1036145291():
    verify_serp_parsing(
        "https://web.archive.org/web/20021101100811id_/http://search.yahoo.com:80/search?p=gamespot+portugal",
        "yahoo",
    )


def test_parse_query_yahoo_gay_shower_sex_961542771():
    verify_serp_parsing(
        "https://web.archive.org/web/20000620231251id_/http://search.yahoo.com:80/search?p=gay+shower+sex",
        "yahoo",
    )


def test_parse_query_yahoo_golf_poll_971193897():
    verify_serp_parsing(
        "https://web.archive.org/web/20001010160457id_/http://search.yahoo.com:80/search?p=golf+poll",
        "yahoo",
    )


def test_parse_query_yahoo_envious_that_biden_1652739375():
    verify_serp_parsing(
        "https://web.archive.org/web/20220516221615id_/https://search.yahoo.com/search?p=envious+that+biden&ei=UTF-8&fr2=time&age=1d&btf=d&fr=sfp",
        "yahoo",
    )


def test_parse_query_yahoo_genserv_971364933():
    verify_serp_parsing(
        "https://web.archive.org/web/20001012153533id_/http://search.yahoo.com:80/search?p=genserv",
        "yahoo",
    )


def test_parse_query_yahoo_every_previewsworld_calls_1649106369():
    verify_serp_parsing(
        "https://web.archive.org/web/20220404210609id_/https://search.yahoo.com/search?p=every+previewsworld+calls&ei=UTF-8&fr2=time&age=1d&btf=d&fr=sfp",
        "yahoo",
    )


def test_parse_query_yahoo_frugal_994624570():
    verify_serp_parsing(
        "https://web.archive.org/web/20010708203610id_/http://search.yahoo.com:80/search?p=Frugal",
        "yahoo",
    )


def test_parse_query_yahoo_hombres_con_pelo_en_el_pecho_photos_1359853955():
    verify_serp_parsing(
        "https://web.archive.org/web/20130203011235id_/http://au.search.yahoo.com/search?p=hombres+con+pelo+en+el+pecho+photos&fr2=sb-top&fr=404&type_param=",
        "yahoo",
    )


def test_parse_query_yahoo_free_pics_bbbw_1041625676():
    verify_serp_parsing(
        "https://web.archive.org/web/20030103202756id_/http://search.yahoo.com:80/search?p=free+pics+bbbw",
        "yahoo",
    )


def test_parse_query_yahoo_funny_970842855():
    verify_serp_parsing(
        "https://web.archive.org/web/20001006143415id_/http://search.yahoo.com:80/search?p=FUNNY",
        "yahoo",
    )


def test_parse_query_yahoo_gender_discrimination_970694307():
    verify_serp_parsing(
        "https://web.archive.org/web/20001004211827id_/http://search.yahoo.com:80/search?p=gender+discrimination+",
        "yahoo",
    )


def test_parse_query_yahoo_gaming_club_997142327():
    verify_serp_parsing(
        "https://web.archive.org/web/20010806235847id_/http://search.yahoo.com:80/search?p=gaming+club",
        "yahoo",
    )


def test_parse_query_yahoo_gordon_bruce_955082521():
    verify_serp_parsing(
        "https://web.archive.org/web/20000407044201id_/http://search.yahoo.com:80/search?p=Gordon+Bruce",
        "yahoo",
    )


def test_parse_query_yahoo_freeware_936293306():
    verify_serp_parsing(
        "https://web.archive.org/web/19990902172826id_/http://search.yahoo.com:80/search?p=freeware",
        "yahoo",
    )


def test_parse_query_yahoo_funny_pictures_943791338():
    verify_serp_parsing(
        "https://web.archive.org/web/19991128121538id_/http://search.yahoo.com:80/search?p=funny+pictures",
        "yahoo",
    )


def test_parse_query_yahoo_free_nude_pics_of_brittney_spears_970473571():
    verify_serp_parsing(
        "https://web.archive.org/web/20001002075931id_/http://search.yahoo.com:80/search?p=free+nude+pics+of+Brittney+Spears",
        "yahoo",
    )


def test_parse_query_yahoo_games_video_games_developers_and_publishers_croteam_1624838167():
    verify_serp_parsing(
        "https://web.archive.org/web/20210627235607id_/https://search.yahoo.com/search?p=Games+Video+Games+Developers+and+Publishers+Croteam",
        "yahoo",
    )


def test_parse_query_yahoo_isnt_even_weirdest_thing_america_right_1653343893():
    verify_serp_parsing(
        "https://web.archive.org/web/20220523221133id_/https://search.yahoo.com/search?p=isnt+even+weirdest+thing+america+right&ei=UTF-8&fr2=time&age=1d&btf=d&fr=sfp",
        "yahoo",
    )


def test_parse_query_yahoo_georgetown_university_law_center_970779220():
    verify_serp_parsing(
        "https://web.archive.org/web/20001005205340id_/http://search.yahoo.com:80/search?p=Georgetown+University+Law+Center",
        "yahoo",
    )


def test_parse_query_yahoo_wheatstone_bridge_1613009822():
    verify_serp_parsing(
        "https://web.archive.org/web/20210211021702id_/https://au.search.yahoo.com/search?p=wheatstone+bridge&fr=my_top",
        "yahoo",
    )


def test_parse_query_yahoo_fresh_prince_of_bel_air_971413040():
    verify_serp_parsing(
        "https://web.archive.org/web/20001013045720id_/http://search.yahoo.com:80/search?p=Fresh+Prince+of+Bel+Air",
        "yahoo",
    )


def test_parse_query_yahoo_mark_schlissel_wife_1642367070():
    verify_serp_parsing(
        "https://web.archive.org/web/20220116210430id_/http://au.search.yahoo.com/search?p=mark+schlissel+wife&ei=UTF-8&nojs=1",
        "yahoo",
    )


def test_parse_query_yahoo_games_puzzles_brain_teasers_sudoku_1574052323():
    verify_serp_parsing(
        "https://web.archive.org/web/20191118044523id_/https://search.yahoo.com/search?p=Games+Puzzles+Brain+Teasers+Sudoku",
        "yahoo",
    )


def test_parse_query_yahoo_gaelic_970435492():
    verify_serp_parsing(
        "https://web.archive.org/web/20001001212452id_/http://search.yahoo.com:80/search?p=Gaelic",
        "yahoo",
    )


def test_parse_query_yahoo_free_xxx_asian_pics_952102259():
    verify_serp_parsing(
        "https://web.archive.org/web/20000303165059id_/http://search.yahoo.com:80/search?p=Free+XXX+asian+pics",
        "yahoo",
    )


def test_parse_query_yahoo_german_pen_pals_961119166():
    verify_serp_parsing(
        "https://web.archive.org/web/20000616013246id_/http://search.yahoo.com:80/search?p=german+pen+pals",
        "yahoo",
    )


def test_parse_query_yahoo_goddess_955975798():
    verify_serp_parsing(
        "https://web.archive.org/web/20000417124958id_/http://search.yahoo.com:80/search?p=Goddess",
        "yahoo",
    )


def test_parse_query_yahoo_govote_com_955100292():
    verify_serp_parsing(
        "https://web.archive.org/web/20000407093812id_/http://search.yahoo.com:80/search?p=govote.com",
        "yahoo",
    )


def test_parse_query_yahoo_trans_texas_corridor_1613259804():
    verify_serp_parsing(
        "https://web.archive.org/web/20210213234324id_/https://au.search.yahoo.com/search?p=trans+texas+corridor&fr=ush-groups",
        "yahoo",
    )


def test_parse_query_yahoo_ghar_1005251824():
    verify_serp_parsing(
        "https://web.archive.org/web/20011108203704id_/http://search.yahoo.com:80/search?p=Ghar",
        "yahoo",
    )


def test_parse_query_yahoo_global_and_parts_970542388():
    verify_serp_parsing(
        "https://web.archive.org/web/20001003030628id_/http://search.yahoo.com:80/search?p=global+and+parts",
        "yahoo",
    )


def test_parse_query_yahoo_georgia_restaurant_association_970314665():
    verify_serp_parsing(
        "https://web.archive.org/web/20000930115105id_/http://search.yahoo.com:80/search?p=Georgia+Restaurant+Association",
        "yahoo",
    )



from pathlib import Path
from typing import MutableMapping
from typing import Sequence, Iterable
from warnings import warn

from click import echo
from click import prompt
from diskcache import Index
from elasticsearch_dsl.query import Terms
from tqdm.auto import tqdm
from whois import whois
from whois.parser import PywhoisError
from yaml import safe_load

from archive_query_log.config import Config
from archive_query_log.orm import Provider
from archive_query_log.parsers.url_offset import add_url_offset_parser
from archive_query_log.parsers.url_page import add_url_page_parser
from archive_query_log.parsers.url_query import add_url_query_parser
from archive_query_log.parsers.warc_query import add_warc_query_parser
from archive_query_log.parsers.warc_snippets import add_warc_snippets_parser
from archive_query_log.parsers.xml import xpaths_from_css_selector, \
    text_xpath, merge_xpaths
from archive_query_log.providers import add_provider
from archive_query_log.utils.es import safe_iter_scan


def _provider_name(
        i: int,
        main_domain: str,
        provider_names: MutableMapping[str, str],
        review: bool,
) -> str | None:
    provider_name_suggest: str | None
    if main_domain in provider_names:
        if not review:
            return provider_names[main_domain]
        else:
            provider_name_suggest = provider_names[main_domain]
    else:
        try:
            main_domain_info = whois(main_domain)
        except PywhoisError:
            main_domain_info = None
        if main_domain_info is not None:
            main_org: str | list[str] | None = main_domain_info.org
            if isinstance(main_org, list):
                main_org = main_org[0]
            if main_org is not None:
                for restricted_phrase in ["redacted", "privacy",
                                          "domain protection", "not disclosed",
                                          "identity protection",
                                          "domains by proxy"]:
                    if restricted_phrase in main_org.casefold():
                        main_org = None
                        break
            if main_org is not None:
                for suffix in ["Inc", "LLC", "Ltd", "LTD", "GmbH", "AG", "S.A",
                               "SE", "Co", "Pty", "B.V", "S.L", "S.R.L",
                               "S.A.S",
                               "SAS", "AB", "&"]:
                    main_org = main_org.removesuffix(f", {suffix}.")
                    main_org = main_org.removesuffix(f", {suffix}")
                    main_org = main_org.removesuffix(f" {suffix}.")
                    main_org = main_org.removesuffix(f" {suffix}")
            provider_name_suggest = main_org
        else:
            provider_name_suggest = None
    provider_name = prompt(
        f"Please enter search provider #{i} name (https://{main_domain})",
        type=str, default=provider_name_suggest,
        show_default=provider_name_suggest != " ")
    if provider_name.strip() == "":
        return None
    provider_names[main_domain] = provider_name
    return provider_name


def import_providers(
        config: Config,
        services_path: Path,
        cache_path: Path,
        review: int | None,
        no_merge: bool,
        auto_merge: bool,
) -> None:
    echo("Load providers from services file.")
    with services_path.open("r") as file:
        services_list: Sequence[dict] = safe_load(file)
    echo(f"Found {len(services_list)} service definitions.")

    services: Iterable[dict] = services_list
    provider_names: MutableMapping[str, str] = Index(str(cache_path))

    if auto_merge or no_merge:
        # noinspection PyTypeChecker
        services = tqdm(
            services,
            desc="Import providers",
            unit="provider",
        )

    num_services = len(services_list)

    ask_for_name = True
    service: dict
    for i, service in enumerate(services):
        if "domains" not in service:
            raise ValueError(f"Service definition #{i} from {services_path} "
                             f"has no domains: {service}")

        if ("query_parsers" not in service or
                len(service["query_parsers"]) == 0):
            continue

        main_domain = service["domains"][0]
        if ask_for_name:
            name = _provider_name(i, main_domain, provider_names,
                                  review is not None and review <= i)
            if name is None:
                ask_for_name = False
        else:
            name = None

        add_provider(
            config=config,
            name=name,
            description=None,
            notes=service.get("notes"),
            exclusion_reason=service.get("excluded"),
            domains=set(service["domains"]),
            url_path_prefixes=set(service["focused_url_prefixes"]),
            priority=num_services - i,
            no_merge=no_merge,
            auto_merge=auto_merge,
        )


def import_url_query_parsers(config: Config, services_path: Path) -> None:
    echo("Load providers from services file.")
    with services_path.open("r") as file:
        services_list: Sequence[dict] = safe_load(file)
    echo(f"Found {len(services_list)} service definitions.")

    services: Iterable[dict] = services_list
    # noinspection PyTypeChecker
    services = tqdm(
        services,
        desc="Import parsers for providers",
        unit="provider",
    )
    for i, service in enumerate(services):
        if "domains" not in service or "query_parsers" not in service:
            continue

        query_parsers = service["query_parsers"]
        num_query_parsers = len(query_parsers)

        providers = (
            Provider.search(using=config.es.client, index=config.es.index_providers)
            .query(Terms(domains=service["domains"]))
            .scan()
        )
        providers = safe_iter_scan(providers)
        for provider in providers:
            for k, query_parser in enumerate(query_parsers):
                if query_parser["type"] == "fragment_segment":
                    warn(UserWarning(
                        f"Service definition #{i} "
                        f"query parser #{k} is of type "
                        f"'fragment_segment', which is not supported."))
                    continue
                remove_patterns = query_parser.get("remove_patterns")
                if remove_patterns is not None:
                    remove_pattern_regex = "|".join(remove_patterns)
                else:
                    remove_pattern_regex = None
                space_patterns = query_parser.get("space_patterns")
                if space_patterns is not None:
                    space_pattern_regex = "|".join(space_patterns)
                else:
                    space_pattern_regex = None
                segment_string = query_parser.get("segment")
                if segment_string is not None:
                    segment = int(segment_string)
                else:
                    segment = None
                add_url_query_parser(
                    config=config,
                    provider_id=provider.meta.id,
                    url_pattern_regex=query_parser.get("url_pattern"),
                    priority=num_query_parsers - k,
                    parser_type=query_parser["type"],
                    parameter=query_parser.get("parameter"),
                    segment=segment,
                    remove_pattern_regex=remove_pattern_regex,
                    space_pattern_regex=space_pattern_regex,
                )


def import_url_page_parsers(config: Config, services_path: Path) -> None:
    echo("Load providers from services file.")
    with services_path.open("r") as file:
        services_list: Sequence[dict] = safe_load(file)
    echo(f"Found {len(services_list)} service definitions.")

    services: Iterable[dict] = services_list
    # noinspection PyTypeChecker
    services = tqdm(
        services,
        desc="Import parsers for providers",
        unit="provider",
    )
    for i, service in enumerate(services):
        if "domains" not in service or "page_parsers" not in service:
            continue

        page_parsers = service["page_parsers"]
        num_page_parsers = len(page_parsers)

        providers = (
            Provider.search(using=config.es.client, index=config.es.index_providers)
            .query(Terms(domains=service["domains"]))
            .scan()
        )
        providers = safe_iter_scan(providers)
        for provider in providers:
            for k, page_parser in enumerate(page_parsers):
                if page_parser["type"] == "fragment_segment":
                    warn(UserWarning(
                        f"Service definition #{i} "
                        f"page parser #{k} is of type "
                        f"'fragment_segment', which is not supported."))
                    continue
                remove_patterns = page_parser.get("remove_patterns")
                if remove_patterns is not None:
                    remove_pattern_regex = "|".join(remove_patterns)
                else:
                    remove_pattern_regex = None
                space_patterns = page_parser.get("space_patterns")
                if space_patterns is not None:
                    space_pattern_regex = "|".join(space_patterns)
                else:
                    space_pattern_regex = None
                segment_string = page_parser.get("segment")
                if segment_string is not None:
                    segment = int(segment_string)
                else:
                    segment = None
                add_url_page_parser(
                    config=config,
                    provider_id=provider.meta.id,
                    url_pattern_regex=page_parser.get("url_pattern"),
                    priority=num_page_parsers - k,
                    parser_type=page_parser["type"],
                    parameter=page_parser.get("parameter"),
                    segment=segment,
                    remove_pattern_regex=remove_pattern_regex,
                    space_pattern_regex=space_pattern_regex,
                )


def import_url_offset_parsers(config: Config, services_path: Path) -> None:
    echo("Load providers from services file.")
    with services_path.open("r") as file:
        services_list: Sequence[dict] = safe_load(file)
    echo(f"Found {len(services_list)} service definitions.")

    services: Iterable[dict] = services_list
    # noinspection PyTypeChecker
    services = tqdm(
        services,
        desc="Import parsers for providers",
        unit="provider",
    )
    for i, service in enumerate(services):
        if "domains" not in service or "offset_parsers" not in service:
            continue

        offset_parsers = service["offset_parsers"]
        num_offset_parsers = len(offset_parsers)

        providers = (
            Provider.search(using=config.es.client, index=config.es.index_providers)
            .query(Terms(domains=service["domains"]))
            .scan()
        )
        providers = safe_iter_scan(providers)
        for provider in providers:
            for k, offset_parser in enumerate(offset_parsers):
                if offset_parser["type"] == "fragment_segment":
                    warn(UserWarning(
                        f"Service definition #{i} "
                        f"offset parser #{k} is of type "
                        f"'fragment_segment', which is not supported."))
                    continue
                remove_patterns = offset_parser.get("remove_patterns")
                if remove_patterns is not None:
                    remove_pattern_regex = "|".join(remove_patterns)
                else:
                    remove_pattern_regex = None
                space_patterns = offset_parser.get("space_patterns")
                if space_patterns is not None:
                    space_pattern_regex = "|".join(space_patterns)
                else:
                    space_pattern_regex = None
                segment_string = offset_parser.get("segment")
                if segment_string is not None:
                    segment = int(segment_string)
                else:
                    segment = None
                add_url_offset_parser(
                    config=config,
                    provider_id=provider.meta.id,
                    url_pattern_regex=offset_parser.get("url_pattern"),
                    priority=num_offset_parsers - k,
                    parser_type=offset_parser["type"],
                    parameter=offset_parser.get("parameter"),
                    segment=segment,
                    remove_pattern_regex=remove_pattern_regex,
                    space_pattern_regex=space_pattern_regex,
                )


def import_warc_query_parsers(config: Config, services_path: Path) -> None:
    echo("Load providers from services file.")
    with services_path.open("r") as file:
        services_list: Sequence[dict] = safe_load(file)
    echo(f"Found {len(services_list)} service definitions.")

    services: Iterable[dict] = services_list
    # noinspection PyTypeChecker
    services = tqdm(
        services,
        desc="Import parsers for providers",
        unit="provider",
    )
    for service in services:
        if ("domains" not in service or
                "interpreted_query_parsers" not in service):
            continue

        interpreted_query_parsers = service["interpreted_query_parsers"]
        num_interpreted_query_parsers = len(interpreted_query_parsers)

        providers = (
            Provider.search(using=config.es.client, index=config.es.index_providers)
            .query(Terms(domains=service["domains"]))
            .scan()
        )
        providers = safe_iter_scan(providers)
        for provider in providers:
            for k, interpreted_query_parser in (
                    enumerate(interpreted_query_parsers)):
                if interpreted_query_parser["type"] != "html_selector":
                    continue
                remove_patterns = (
                    interpreted_query_parser.get("remove_patterns"))
                if remove_patterns is not None:
                    remove_pattern_regex = "|".join(remove_patterns)
                else:
                    remove_pattern_regex = None
                space_patterns = interpreted_query_parser.get("space_patterns")
                if space_patterns is not None:
                    space_pattern_regex = "|".join(space_patterns)
                else:
                    space_pattern_regex = None
                query_selector = interpreted_query_parser["query_selector"]

                query_text = interpreted_query_parser.get("query_text", False)
                query_attribute = interpreted_query_parser.get(
                    "query_attribute", "value" if not query_text else None)

                query_xpaths = xpaths_from_css_selector(query_selector)
                query_xpaths = [
                    "//" +
                    text_xpath(
                        query_xpath,
                        attribute=query_attribute,
                        text=query_text,
                    )
                    for query_xpath in query_xpaths
                ]
                query_xpath = merge_xpaths(query_xpaths)

                add_warc_query_parser(
                    config=config,
                    provider_id=provider.meta.id,
                    url_pattern_regex=interpreted_query_parser.get(
                        "url_pattern"),
                    priority=num_interpreted_query_parsers - k,
                    parser_type="xpath",
                    xpath=query_xpath,
                    remove_pattern_regex=remove_pattern_regex,
                    space_pattern_regex=space_pattern_regex,
                )


def import_warc_snippets_parsers(config: Config, services_path: Path) -> None:
    echo("Load providers from services file.")
    with services_path.open("r") as file:
        services_list: Sequence[dict] = safe_load(file)
    echo(f"Found {len(services_list)} service definitions.")

    services: Iterable[dict] = services_list
    # noinspection PyTypeChecker
    services = tqdm(
        services,
        desc="Import parsers for providers",
        unit="provider",
    )
    for service in services:
        if ("domains" not in service or "results_parsers" not in service):
            continue

        results_parsers = service["results_parsers"]
        num_results_parsers = len(results_parsers)

        providers = (
            Provider.search(using=config.es.client, index=config.es.index_providers)
            .query(Terms(domains=service["domains"]))
            .scan()
        )
        providers = safe_iter_scan(providers)
        for provider in providers:
            for k, results_parser in enumerate(results_parsers):
                if results_parser["type"] != "html_selector":
                    continue
                results_selector = results_parser["results_selector"]
                url_selector = results_parser.get("url_selector")
                title_selector = results_parser.get("title_selector")
                snippet_selector = results_parser.get("snippet_selector")

                results_xpaths = xpaths_from_css_selector(results_selector)
                results_xpaths = [
                    "//" + result_xpath
                    for result_xpath in results_xpaths
                ]
                results_xpath = merge_xpaths(results_xpaths)

                if url_selector is not None:
                    url_xpaths = xpaths_from_css_selector(url_selector)
                    url_xpaths = [
                        text_xpath(xpath, attribute="href")
                        for xpath in url_xpaths
                    ]
                    url_xpath = merge_xpaths(url_xpaths)
                else:
                    url_xpath = None

                if title_selector is not None:
                    title_xpaths = xpaths_from_css_selector(title_selector)
                    title_xpaths = [
                        text_xpath(xpath, text=True)
                        for xpath in title_xpaths
                    ]
                    title_xpath = merge_xpaths(title_xpaths)
                else:
                    title_xpath = None

                if snippet_selector is not None:
                    snippet_xpaths = xpaths_from_css_selector(snippet_selector)
                    snippet_xpaths = [
                        text_xpath(xpath, text=True)
                        for xpath in snippet_xpaths
                    ]
                    snippet_xpath = merge_xpaths(snippet_xpaths)
                else:
                    snippet_xpath = None

                add_warc_snippets_parser(
                    config=config,
                    provider_id=provider.meta.id,
                    url_pattern_regex=results_parser.get("url_pattern"),
                    priority=num_results_parsers - k,
                    parser_type="xpath",
                    xpath=results_xpath,
                    url_xpath=url_xpath,
                    title_xpath=title_xpath,
                    text_xpath=snippet_xpath,
                )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_indeed_origins_macy_s_1353165507():
    verify_serp_parsing(
        "https://web.archive.org/web/20121117151827id_/http://www.indeed.com/jobs?q=Origins%2Fmacy%27s",
        "indeed",
    )


def test_parse_query_indeed_hca_corporate_1334425152():
    verify_serp_parsing(
        "https://web.archive.org/web/20120414173912id_/http://www.indeed.com/jobs?q=HCA-Corporate&l=Orlando,+FL",
        "indeed",
    )


def test_parse_query_indeed_60000_1450529560():
    verify_serp_parsing(
        "https://web.archive.org/web/20151219125240id_/http://www.indeed.com/jobs?q=$60,000&l=Saratoga,+CA&sort=date",
        "indeed",
    )


def test_parse_query_indeed_concept_development_integration_1647187061():
    verify_serp_parsing(
        "https://web.archive.org/web/20220313155741id_/https://de.indeed.com/jobs?q=Concept+Development+Integration&l=M%C3%BCnchen",
        "indeed",
    )


def test_parse_query_indeed_jvm_lending_1586272513():
    verify_serp_parsing(
        "https://web.archive.org/web/20200407151513id_/https://www.indeed.com/jobs?q=JVM+Lending&rbl=Salt+Lake+City,+UT&jlid=029db43e55e0cfca",
        "indeed",
    )


def test_parse_query_indeed_customer_service_1447983624():
    verify_serp_parsing(
        "https://web.archive.org/web/20151120014024id_/http://www.indeed.com/jobs?q=Customer+Service&rbl=Phoenix,+AZ&jlid=b4ff2ac66b411fd8",
        "indeed",
    )


def test_parse_query_indeed_higher_education_freelance_network_1585036169():
    verify_serp_parsing(
        "https://web.archive.org/web/20200324074929id_/https://www.indeed.com/jobs?q=Higher+Education+Freelance+Network&rbc=The+University+of+Chicago&jcid=5cb727313a823dd6",
        "indeed",
    )


def test_parse_query_indeed_steiner_business_solutions_1577184429():
    verify_serp_parsing(
        "https://web.archive.org/web/20191224104709id_/https://www.indeed.com/jobs?q=Steiner+Business+Solutions&explvl=mid_level",
        "indeed",
    )


def test_parse_query_indeed_international_english_prep_academy_iepa_1586237411():
    verify_serp_parsing(
        "https://web.archive.org/web/20200407053011id_/https://www.indeed.com/jobs?q=International+English+Prep+Academy+(iepa)&l=Stafford,+VA&nc=jasx",
        "indeed",
    )


def test_parse_query_indeed_pepsico_1487893385():
    verify_serp_parsing(
        "https://web.archive.org/web/20170223234305id_/http://www.indeed.com/jobs?q=PepsiCo&jt=contract",
        "indeed",
    )



from datetime import datetime, timezone, timedelta

EPOCH = datetime.fromtimestamp(0)

UTC = timezone.utc

CET = timezone(timedelta(hours=1))
"""
Central European Time (CET)
"""


def utc_now() -> datetime:
    return datetime.now(tz=UTC).replace(microsecond=0)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_roblox_survival_1656759229():
    verify_serp_parsing(
        "https://web.archive.org/web/20220702105349id_/https://www.roblox.com/discover/?Keyword=survival",
        "roblox",
    )


def test_parse_query_roblox_survive_the_killer_1640549700():
    verify_serp_parsing(
        "https://web.archive.org/web/20211226201500id_/https://www.roblox.com/discover/?Keyword=survive+the+killer",
        "roblox",
    )


def test_parse_query_roblox_chernobyl_rp_1666555966():
    verify_serp_parsing(
        "https://web.archive.org/web/20221023201246id_/https://www.roblox.com/discover/?Keyword=chernobyl%20rp",
        "roblox",
    )


def test_parse_query_roblox_shorts_1357668619():
    verify_serp_parsing(
        "https://web.archive.org/web/20130108181019id_/http://www.roblox.com:80/catalog/browse.aspx?Subcategory=13&Keyword=shorts&LegendExpanded=true&Category=3",
        "roblox",
    )


def test_parse_query_roblox_xo_so_88_gg8_run_1659417607():
    verify_serp_parsing(
        "https://web.archive.org/web/20220802052007id_/https://www.roblox.com/discover/?Keyword=x%E1%BB%95%20s%E1%BB%91%2088%20GG8.Run",
        "roblox",
    )


def test_parse_query_roblox_welcome_to_the_town_of_robloxia_uncopylocked_1660007534():
    verify_serp_parsing(
        "https://web.archive.org/web/20220809011214id_/https://www.roblox.com/discover/?Keyword=welcome%20to%20the%20town%20of%20robloxia%20uncopylocked",
        "roblox",
    )


def test_parse_query_roblox_build_to_survive_black_people_1663785067():
    verify_serp_parsing(
        "https://web.archive.org/web/20220921183107id_/https://www.roblox.com/discover/?Keyword=build%20to%20survive%20black%20people",
        "roblox",
    )


def test_parse_query_roblox_cut_1633682743():
    verify_serp_parsing(
        "https://web.archive.org/web/20211008084543id_/https://www.roblox.com/discover/?Keyword=cut",
        "roblox",
    )


def test_parse_query_roblox_znation8000_1650350926():
    verify_serp_parsing(
        "https://web.archive.org/web/20220419064846id_/https://www.roblox.com/discover/?Keyword=znation8000",
        "roblox",
    )


def test_parse_query_roblox_soviet_union_1657601847():
    verify_serp_parsing(
        "https://web.archive.org/web/20220712045727id_/https://www.roblox.com/discover/?Keyword=soviet+union",
        "roblox",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_canva_instagram_reels_video_1607594697():
    verify_serp_parsing(
        "https://web.archive.org/web/20201210100457id_/https://www.canva.com/design/play?category=tAEEZ_gIPxM&template=EAD_mTZpt2k&type=TAD9VKDO5tI&layoutQuery=Instagram%20Reels%20Video&analyticsCorrelationId=f6767fc3-7f33-445e-9740-a004311e5afe",
        "canva",
    )



from dataclasses import dataclass
from functools import cached_property
from gzip import open as gzip_open, GzipFile
from json import JSONDecodeError
from pathlib import Path
from typing import Sized, Iterable, Iterator

from marshmallow import Schema
from warcio.archiveiterator import ArchiveIterator
from warcio.recordloader import ArcWarcRecord

from archive_query_log.legacy import LOGGER
from archive_query_log.legacy.model import ArchivedQueryUrl, ArchivedRawSerp


@dataclass(frozen=True)
class ArchivedRawSerps(Sized, Iterable[ArchivedRawSerp]):
    """
    Read archived raw SERP contents from a directory of WARC files.
    """

    path: Path
    """
    Path where the raw SERP contents are stored in WARC format.
    """

    def __post_init__(self):
        self._check_raw_serps_paths()

    def _check_raw_serps_paths(self):
        if not self.path.exists() or not self.path.is_dir():
            raise ValueError(
                f"Raw SERPs path must be a directory: {self.path}"
            )

    def _streams(self) -> Iterator[tuple[Path, GzipFile]]:
        files = self.path.glob("*.warc.gz")
        for file in files:
            with gzip_open(file, "rb") as stream:
                yield file, stream

    def __len__(self) -> int:
        return sum(
            1
            for _, stream in self._streams()
            for record in ArchiveIterator(stream, no_record_parse=True)
            if record.rec_type == "response"
        )

    @cached_property
    def _archived_serp_url_schema(self) -> Schema:
        return ArchivedQueryUrl.schema()

    def _read_serp_content(
            self,
            record: ArcWarcRecord,
    ) -> ArchivedRawSerp | None:
        archived_serp_url: ArchivedQueryUrl
        record_url_header = record.rec_headers.get_header("Archived-URL")
        try:
            archived_serp_url = self._archived_serp_url_schema.loads(
                record_url_header
            )
        except JSONDecodeError:
            LOGGER.warning(
                f"Could not index {record_url_header} from record "
                f"{record.rec_headers.get_header('WARC-Record-ID')}."
            )
            return None
        encoding = record.http_headers.get_header("Content-Type")
        if encoding is None:
            encoding = ""
        encoding = encoding.split(";")[-1].split("=")[-1].strip().lower()
        if encoding == "" or "/" in encoding:
            encoding = "utf8"
        return ArchivedRawSerp(
            url=archived_serp_url.url,
            timestamp=archived_serp_url.timestamp,
            query=archived_serp_url.query,
            page=archived_serp_url.page,
            offset=archived_serp_url.offset,
            content=record.content_stream().read(),
            encoding=encoding,
        )

    def __iter__(self) -> Iterator[ArchivedRawSerp]:
        for path, stream in self._streams():
            failures = False
            for record in ArchiveIterator(stream):
                if record.rec_type != "response":
                    continue
                serp = self._read_serp_content(record)
                if serp is None:
                    failures = True
                    continue
                yield serp
            if failures:
                LOGGER.warning(f"Failed to index some records from {path}.")



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_espn_ball_state_1619440827():
    verify_serp_parsing(
        "https://web.archive.org/web/20210426124027id_/https://www.espn.com/search/_/q/Ball%20State/o/watch",
        "espn",
    )


def test_parse_query_espn_ball_state_1615730301():
    verify_serp_parsing(
        "https://web.archive.org/web/20210314135821id_/https://www.espn.com/search/_/q/Ball%20State/o/watch",
        "espn",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_sogou_lofke_yi_kua_shi_chang_mai_mai_1333092705():
    verify_serp_parsing(
        "https://web.archive.org/web/20120330073145id_/http://www.sogou.com/web?query=lof%E5%8F%AF%E4%BB%A5%E8%B7%A8%E5%B8%82%E5%9C%BA%E4%B9%B0%E5%8D%96&p=02210102&fhintidx=7",
        "sogou",
    )


def test_parse_query_sogou_xing_ai_ji_qiao_1576522689():
    verify_serp_parsing(
        "https://web.archive.org/web/20191216185809id_/https://www.sogou.com/sogou?query=%E6%80%A7%E7%88%B1%E6%8A%80%E5%B7%A7&ie=utf8&insite=baike.sogou.com&pid=sogou-wsse-17737832ac17be52",
        "sogou",
    )


def test_parse_query_sogou_j_1341637724():
    verify_serp_parsing(
        "https://web.archive.org/web/20120707050844id_/http://www.sogou.com:80/web?query=%D7%F6%BF%D5%BB%FA%B9%B9%CF%E3%E9%DA&p=31210100&fhintidx=3",
        "sogou",
    )


def test_parse_query_sogou_ju_min_shen_fen_zheng_hao_ma_he_xing_ming_1508756794():
    verify_serp_parsing(
        "https://web.archive.org/web/20171023110634id_/http://www.sogou.com/sogou?user_ip=207.241.229.48&sourceid=hint&bh=1&hintidx=11&query=%E5%B1%85%E6%B0%91%E8%BA%AB%E4%BB%BD%E8%AF%81%E5%8F%B7%E7%A0%81%E5%92%8C%E5%A7%93%E5%90%8D&pid=sogou-site-d54ce9de9df77c57&duppid=1&w=01020600&interation=&interV=kKIOkrELjboLmLkEkLoTkKIMkrELjboImLkEk74TkKILmrELjb8TkKIKmrELjbkI_-1099283536&htdbg=idc%3Ebjdjt%7CdbgID%3E01%7Cabt%3E0%7Cmth%3E1&ie=utf8&s_from=hint_last",
        "sogou",
    )


def test_parse_query_sogou_yuan_dai_ma_xie_lu_1578201651():
    verify_serp_parsing(
        "https://web.archive.org/web/20200105052051id_/https://www.sogou.com/sogou?pid=sogou-wsse-ff111e4a5406ed40&insite=zhihu.com&p=42351201&query=%E6%BA%90%E4%BB%A3%E7%A0%81%E6%B3%84%E9%9C%B2&ie=utf8&p=42351201&query=%E6%BA%90%E4%BB%A3%E7%A0%81%E6%B3%84%E9%9C%B2&ie=utf8",
        "sogou",
    )


def test_parse_query_sogou_cesuk_163163163_cn_1493482889():
    verify_serp_parsing(
        "https://web.archive.org/web/20170429162129id_/http://www.sogou.com/sogou?pid=sogou-wsse-3f7bcd0b3ea82268&ie=utf-8&query=cesuk.163163163.cn",
        "sogou",
    )


def test_parse_query_sogou_l_1346233371():
    verify_serp_parsing(
        "https://web.archive.org/web/20120829094251id_/http://www.sogou.com:80/web?query=%BF%F7%CB%F0%B9%FD%B5%C4%B9%CC%B6%A8%CA%D5%D2%E6%D0%C5%CD%D0&p=02210102&fhintidx=7",
        "sogou",
    )


def test_parse_query_sogou_eed_1_4a1_a_1332923188():
    verify_serp_parsing(
        "https://web.archive.org/web/20120328082628id_/http://www.sogou.com/web?query=%C2%B1%C3%88%C3%88%C3%B0%C2%BC%C2%AA%C2%B9%C2%B7%C3%81%C2%B8&p=03210100",
        "sogou",
    )


def test_parse_query_sogou_ti_gong_jin_kou_qi_qiang_1647324224():
    verify_serp_parsing(
        "https://web.archive.org/web/20220315060344id_/https://www.sogou.com/web?query=%E6%8F%90%E4%BE%9B%E8%BF%9B%E5%8F%A3%E6%B0%94%E6%9E%AA&ie=utf8",
        "sogou",
    )


def test_parse_query_sogou_tu_guan_xin_gai_kuan_1332227339():
    verify_serp_parsing(
        "https://web.archive.org/web/20120320070859id_/http://www.sogou.com/web?query=%E9%80%94%E8%A7%82%E6%96%B0%E6%94%B9%E6%AC%BE&p=02210102&fhintidx=3",
        "sogou",
    )



from uuid import uuid5, NAMESPACE_URL

NAMESPACE_AQL = uuid5(NAMESPACE_URL, "aql")
NAMESPACE_SOURCE = uuid5(NAMESPACE_AQL, "filter")
NAMESPACE_CAPTURE = uuid5(NAMESPACE_AQL, "capture")
NAMESPACE_SERP = uuid5(NAMESPACE_AQL, "serp")
NAMESPACE_RESULT = uuid5(NAMESPACE_AQL, "result")
NAMESPACE_URL_QUERY_PARSER = uuid5(NAMESPACE_AQL, "url_query_parser")
NAMESPACE_URL_PAGE_PARSER = uuid5(NAMESPACE_AQL, "url_page_parser")
NAMESPACE_URL_OFFSET_PARSER = uuid5(NAMESPACE_AQL, "url_offset_parser")
NAMESPACE_URL_LANGUAGE_PARSER = uuid5(
    NAMESPACE_AQL, "url_language_parser")
NAMESPACE_WARC_QUERY_PARSER = uuid5(NAMESPACE_AQL, "warc_query_parser")
NAMESPACE_WARC_SNIPPETS_PARSER = uuid5(
    NAMESPACE_AQL, "warc_snippets_parser")
NAMESPACE_WARC_MAIN_CONTENT_PARSER = uuid5(
    NAMESPACE_AQL, "warc_main_content_parser")
NAMESPACE_WARC_DIRECT_ANSWERS_PARSER = uuid5(
    NAMESPACE_AQL, "warc_direct_answers_parser")
NAMESPACE_WARC_DOWNLOADER = uuid5(NAMESPACE_AQL, "warc_downloader")



from dataclasses import dataclass
from datetime import datetime
from itertools import chain, islice
from json import JSONEncoder, JSONDecoder
from pathlib import Path
from re import (
    compile as re_compile,
    VERBOSE as RE_VERBOSE,
    MULTILINE as RE_MULTILINE,
    DOTALL as RE_DOTALL,
)
from typing import Iterable, Iterator, TypeVar, Generic, Type, Callable, Any
from uuid import uuid5
from warnings import warn

from click import echo
from elasticsearch_dsl import Document, Search
from elasticsearch_dsl.function import RandomScore
from elasticsearch_dsl.query import FunctionScore, Term, RankFeature

# from elasticsearch_dsl.query import Exists
from requests import ConnectionError as RequestsConnectionError
from tqdm.auto import tqdm
from warc_cache import WarcCacheStore, WarcCacheRecord
from warc_s3 import WarcS3Store
from warcio.recordloader import ArcWarcRecord as WarcRecord

# from web_archive_api.cdx import CdxApi, CdxMatchType, CdxCapture
from web_archive_api.memento import MementoApi

from archive_query_log import __version__ as app_version
from archive_query_log.config import Config
from archive_query_log.namespaces import NAMESPACE_WARC_DOWNLOADER
from archive_query_log.orm import Serp, InnerDownloader, WarcLocation

# from archive_query_log.orm import Result
from archive_query_log.utils.es import safe_iter_scan, update_action
from archive_query_log.utils.time import utc_now

_T = TypeVar("_T", bound=Document)


_PATTERN_ISO_FORMAT = re_compile(
    r"\d{4}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d\.\d+([+-][0-2]\d:[0-5]\d|Z)"
)

_PATTERN_WHITESPACE = re_compile(
    r"[ \t\n\r]*", flags=RE_VERBOSE | RE_MULTILINE | RE_DOTALL
)


class _JsonEncoder(JSONEncoder):
    def default(self, object: Any) -> str | Any:
        if isinstance(object, datetime):
            return object.isoformat()
        return super().default(object)


_JSON_ENCODER = _JsonEncoder()


class _JsonDecoder(JSONDecoder):
    def _decode_isoformat(self, object: Any) -> Any:
        if isinstance(object, str) and _PATTERN_ISO_FORMAT.match(object):
            return datetime.fromisoformat(object)
        if isinstance(object, list):
            return [self._decode_isoformat(value) for value in object]
        if isinstance(object, dict):
            return {key: self._decode_isoformat(value) for key, value in object.items()}
        return object

    def decode(self, string: str, _w=_PATTERN_WHITESPACE.match) -> Any:
        obj = super().decode(string, _w)
        obj = self._decode_isoformat(obj)
        return obj


_JSON_DECODER = _JsonDecoder()


class _WrapperWarcRecord(WarcRecord, Generic[_T]):
    _wrapped_type: Type[_T]

    def __init__(self, record: WarcRecord, wrapped: _T | Type[_T]) -> None:
        super().__init__(
            record.format,
            record.rec_type,
            record.rec_headers,
            record.raw_stream,
            record.http_headers,
            record.content_type,
            record.length,
            payload_length=record.payload_length,
            digest_checker=record.digest_checker,
        )
        if isinstance(wrapped, type):
            self._wrapped_type = wrapped
        else:
            self._wrapped_type = type(wrapped)
            self.rec_headers["WARC-Wrapped"] = _JSON_ENCODER.encode(
                wrapped.to_dict(include_meta=True)
            )

    @property
    def wrapped(self) -> _T:
        data = _JSON_DECODER.decode(self.rec_headers["WARC-Wrapped"])
        wrapped = self._wrapped_type.from_es(data)
        return wrapped


class _AnnotatedWarcRecord(WarcRecord, Generic[_T]):
    annotation: _T

    def __init__(self, record: WarcRecord, annotation: _T) -> None:
        super().__init__(
            record.format,
            record.rec_type,
            record.rec_headers,
            record.raw_stream,
            record.http_headers,
            record.content_type,
            record.length,
            payload_length=record.payload_length,
            digest_checker=record.digest_checker,
        )
        self.annotation = annotation


class _SerpWrapperWarcRecord(_WrapperWarcRecord[Serp]):
    pass


def _download_serp_warc(
    config: Config,
    serp: Serp,
) -> Iterable[_SerpWrapperWarcRecord]:
    if serp.capture.status_code != 200:
        return
    memento_api = MementoApi(
        api_url=serp.archive.memento_api_url,
        session=config.http.session,
    )
    try:
        records = memento_api.load_url_warc(
            url=serp.capture.url,
            timestamp=serp.capture.timestamp,
            raw=True,
        )
    except RequestsConnectionError:
        warn(
            RuntimeWarning(
                f"Connection error while downloading WARC "
                f"for capture URL {serp.capture.url} at {serp.capture.timestamp}."
            )
        )
        return
    for record in records:
        yield _SerpWrapperWarcRecord(record, serp)


def download_serps_warc(config: Config, prefetch_limit: int | None = None) -> None:
    changed_serps_search: Search = (
        Serp.search(using=config.es.client, index=config.es.index_serps)
        .filter(
            Term(capture__status_code=200)
            & ~Term(warc_downloader__should_download=False)
            # FIXME: Remove this manual prioritization at some time.
            & Term(provider__id="f205fc44-d918-4b79-9a7f-c1373a6ff9f2")
        )
        .query(
            RankFeature(field="archive.priority", saturation={})
            | RankFeature(field="provider.priority", saturation={})
            | FunctionScore(functions=[RandomScore()])
        )
    )
    num_changed_serps = changed_serps_search.count()

    if num_changed_serps <= 0:
        echo("No new/changed SERPs.")
        return

    changed_serps: Iterable[Serp] = changed_serps_search.params(
        preserve_order=True
    ).scan()
    changed_serps = safe_iter_scan(changed_serps)

    if prefetch_limit is not None:
        num_changed_serps = min(num_changed_serps, prefetch_limit)
        changed_serps = tqdm(changed_serps, total=num_changed_serps, desc="Pre-fetching SERPs", unit="SERP")
        changed_serps = iter(list(islice(changed_serps, prefetch_limit)))
        
    # noinspection PyTypeChecker
    changed_serps = tqdm(
        changed_serps, total=num_changed_serps, desc="Downloading WARCs", unit="SERP"
    )

    # Download from Memento API.
    serp_records: Iterable[_SerpWrapperWarcRecord] = chain.from_iterable(
        _download_serp_warc(config, serp) for serp in changed_serps
    )

    # Write to cache.
    locations: Iterator[WarcCacheRecord] = config.warc_cache.store_serps.write(
        serp_records
    )
    # Consume iterator to write to cache.
    for _ in locations:
        pass


_R = TypeVar("_R", bound=WarcRecord)


@dataclass(frozen=True)
class _WithClearCallback(Generic[_T]):
    payload: _T
    clear: Callable[[], None]


def _iter_cached_records(
    warc_store: WarcCacheStore,
) -> Iterator[_WithClearCallback[WarcRecord]]:
    """
    Re-iterate the cached records from the WARC cache and keep track of the cache files that were completely read.
    A clear callback is provided to remove the completely read cache files.
    """

    completed_paths: list[Path] = []

    def _clear() -> None:
        if len(completed_paths) > 0:
            print(f"Clearing {len(completed_paths)} cached files.")
        while len(completed_paths) > 0:
            path = completed_paths[0]
            path.unlink()
            completed_paths.remove(path)

    last_path: Path | None = None
    for cache_record in warc_store.read_all():
        record = cache_record.record

        path = warc_store.cache_dir_path / cache_record.location.key

        yield _WithClearCallback(record, _clear)

        if last_path is not None and last_path != path:
            completed_paths.append(last_path)
        last_path = path

    # Ensure last cache file is added to completed paths after all records have been iterated.
    if last_path is not None:
        completed_paths.append(last_path)


def _iter_wrapped_records(
    records: Iterable[_WithClearCallback[WarcRecord]],
    wrapped_type: Type[_T],
) -> Iterator[_WithClearCallback[_WrapperWarcRecord[_T]]]:
    """
    Interpret the WARC records as wrapped records with a payload of a specific document type in its WARC-Wrapped header.
    For example, a WARC record might contain a SERP document in its WARC-Wrapped header, which we later use to update the corresponding SERP on Elasticsearch.
    """

    for record_with_callback in records:
        yield _WithClearCallback(
            payload=_WrapperWarcRecord(
                record=record_with_callback.payload,
                wrapped=wrapped_type,
            ),
            clear=record_with_callback.clear,
        )


def _iter_annotated_records(
    records: Iterable[_WithClearCallback[_WrapperWarcRecord[_T]]],
) -> Iterator[_AnnotatedWarcRecord[_WithClearCallback[_T]]]:
    """
    Convert the wrapped records (with a visible WARC header) to "annotated" records (where the payload is opaque to the actual WARC record).
    This is useful for storing the records in S3, where the WARC-Wrapped header should be removed. But still, we need to keep track of the payload for later use.
    """

    for record in records:
        document = record.payload.wrapped
        del record.payload.rec_headers["WARC-Wrapped"]

        yield _AnnotatedWarcRecord(
            record=record.payload,
            annotation=_WithClearCallback(
                payload=document,
                clear=record.clear,
            ),
        )


def _iter_s3_stored_records(
    records: Iterable[_AnnotatedWarcRecord[_WithClearCallback[_T]]],
    warc_store: WarcS3Store,
    document_type: Type[_T],
) -> Iterator[tuple[_T, WarcLocation]]:
    """
    Store the annotated records in S3 and yield the stored documents with their corresponding locations on S3.
    """

    stored_records = warc_store.write(records)

    last_key: str | None = None
    last_clear: Callable[[], None] | None = None
    for stored_record in stored_records:
        location = WarcLocation(
            file=stored_record.location.key,
            offset=stored_record.location.offset,
            length=stored_record.location.length,
        )

        record = stored_record.record
        if not isinstance(record, _AnnotatedWarcRecord):
            raise TypeError(f"Expected {_AnnotatedWarcRecord}, got {type(record)}.")

        annotation = record.annotation
        if not isinstance(annotation, _WithClearCallback):
            raise TypeError(f"Expected _WithClearCallback, got {type(annotation)}.")

        document = annotation.payload
        clear = annotation.clear
        if not isinstance(document, document_type):
            raise TypeError(f"Expected {document_type}, got {type(document)}.")

        yield document, location

        # Clear cache file after storing in S3.
        current_key = stored_record.location.key
        if last_key is None:
            last_key = current_key
        elif last_key != current_key:
            clear()
            last_key = current_key
            last_clear = None
        last_clear = clear

    # Ensure last cache file is cleared after all records have been iterated.
    if last_clear is not None:
        last_clear()


def upload_serps_warc(config: Config) -> None:
    # Read from cache.
    cached_records = _iter_cached_records(warc_store=config.warc_cache.store_serps)

    # Parse wrapped records (document visible in a WARC header).
    wrapped_records = _iter_wrapped_records(
        records=cached_records,
        wrapped_type=Serp,
    )

    # Transform to annotated records (document opaque to the actual WARC record).
    annotated_records = _iter_annotated_records(
        records=wrapped_records,
    )

    # Write to S3.
    stored_serps = _iter_s3_stored_records(
        records=annotated_records,
        warc_store=config.s3.warc_store,
        document_type=Serp,
    )

    # Get downloader ID.
    downloader_id_components = (
        config.s3.endpoint_url,
        config.s3.bucket_name,
        app_version,
    )
    downloader_id = str(
        uuid5(
            NAMESPACE_WARC_DOWNLOADER,
            ":".join(downloader_id_components),
        )
    )

    # Update Elasticsearch.
    actions = (
        update_action(
            document=serp,
            warc_location=location,
            warc_downloader=InnerDownloader(
                id=downloader_id,
                should_download=False,
                last_downloaded=utc_now(),
            ),
        )
        for serp, location in stored_serps
    )

    config.es.bulk(actions)


# class _ResultArcWarcRecord(_WrapperWarcRecord[Result]):
#     pass


# def _capture_timestamp_distance(timestamp: datetime) -> Callable[[CdxCapture], float]:
#     def _distance(capture: CdxCapture) -> float:
#         return abs(timestamp - capture.timestamp).total_seconds()

#     return _distance


# def _download_result_warc(
#     config: Config,
#     result: Result,
# ) -> Iterator[_ResultArcWarcRecord]:
#     if result.snippet.url is None:
#         return

#     cdx_api = CdxApi(
#         api_url=result.archive.cdx_api_url,
#         session=config.http.session,
#     )
#     memento_api = MementoApi(
#         api_url=result.archive.memento_api_url,
#         session=config.http.session,
#     )

#     capture_timestamp = result.capture.timestamp
#     nearest_result_capture_before_serp: CdxCapture | None = min(
#         cdx_api.iter_captures(
#             result.snippet.url,
#             match_type=CdxMatchType.EXACT,
#             to_timestamp=capture_timestamp,
#         ),
#         key=_capture_timestamp_distance(capture_timestamp),
#         default=None,
#     )
#     nearest_result_capture_after_serp: CdxCapture | None = min(
#         cdx_api.iter_captures(
#             result.snippet.url,
#             match_type=CdxMatchType.EXACT,
#             from_timestamp=capture_timestamp,
#         ),
#         key=_capture_timestamp_distance(capture_timestamp),
#         default=None,
#     )
#     if nearest_result_capture_before_serp is None:
#         result.update(
#             using=config.es.client,
#             index=config.es.index_results,
#             warc_before_serp_downloader=InnerDownloader(
#                 should_download=False,
#                 last_downloaded=utc_now(),
#             ).to_dict(),
#         )
#     else:
#         records = memento_api.load_capture_warc(
#             capture=nearest_result_capture_before_serp,
#             raw=True,
#         )
#         for record in records:
#             yield _ResultArcWarcRecord(result, record)
#     if nearest_result_capture_after_serp is None:
#         result.update(
#             using=config.es.client,
#             index=config.es.index_results,
#             warc_after_serp_downloader=InnerDownloader(
#                 should_download=False,
#                 last_downloaded=utc_now(),
#             ).to_dict(),
#         )
#     else:
#         records = memento_api.load_capture_warc(
#             capture=nearest_result_capture_after_serp,
#             raw=True,
#         )
#         for record in records:
#             yield _ResultArcWarcRecord(result, record)


# def download_results_warc(config: Config) -> None:
#     changed_results_search: Search = (
#         Result.search(using=config.es.client, index=config.es.index_results)
#         .filter(Exists(field="snippet.url") & ~Term(should_fetch_captures=False))
#         .query(
#             RankFeature(field="archive.priority", saturation={})
#             | RankFeature(field="provider.priority", saturation={})
#             | FunctionScore(functions=[RandomScore()])
#         )
#     )
#     num_changed_results = changed_results_search.count()

#     if num_changed_results <= 0:
#         echo("No new/changed results.")
#         return

#     changed_results: Iterable[Result] = changed_results_search.scan()
#     changed_results = safe_iter_scan(changed_results)
#
#     if prefetch_limit is not None:
#         num_changed_results = min(num_changed_results, prefetch_limit)
#         changed_results = tqdm(changed_results, total=num_changed_results, desc="Pre-fetching results", unit="result")
#         changed_results = iter(list(islice(changed_results, prefetch_limit)))
#
#     # noinspection PyTypeChecker
#     changed_results = tqdm(
#         changed_results,
#         total=num_changed_results,
#         desc="Downloading WARCs",
#         unit="result",
#     )

#     # Download from Memento API.
#     result_records = chain.from_iterable(
#         _download_result_warc(config, result) for result in changed_results
#     )

#     # Write to S3.
#     stored_records: Iterator[WarcS3Record] = config.s3.warc_store.write(result_records)
#     stored_results = (
#         _unwrap(record, _ResultArcWarcRecord) for record in stored_records
#     )

#     downloader_id_components = (
#         config.s3.endpoint_url,
#         config.s3.bucket_name,
#         app_version,
#     )
#     downloader_id = str(
#         uuid5(
#             NAMESPACE_WARC_DOWNLOADER,
#             ":".join(downloader_id_components),
#         )
#     )
#     actions = (
#         update_action(
#             result,
#             warc_location=location,
#             warc_downloader=InnerDownloader(
#                 id=downloader_id,
#                 should_download=False,
#                 last_downloaded=utc_now(),
#             ),
#         )
#         for result, location in stored_results
#     )
#     config.es.bulk(actions)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_csdn_fifo_1663204137():
    verify_serp_parsing(
        "https://web.archive.org/web/20220915010857id_/https://so.csdn.net/so/search?q=FIFO&spm=1001.2101.3001.7020",
        "csdn",
    )


def test_parse_query_csdn_rows_1665572662():
    verify_serp_parsing(
        "https://web.archive.org/web/20221012110422id_/https://so.csdn.net/so/search?q=rows&spm=1001.2101.3001.7020",
        "csdn",
    )


def test_parse_query_csdn_android_studio_1660487187():
    verify_serp_parsing(
        "https://web.archive.org/web/20220814142627id_/https://so.csdn.net/so/search?q=android%20studio",
        "csdn",
    )



from pathlib import Path

from click import group, Path as PathType, argument, option

from archive_query_log.cli.util import pass_config
from archive_query_log.config import Config
from archive_query_log.orm import Capture


@group()
def captures() -> None:
    pass


@captures.command()
@pass_config
@option(
    "--prefetch-limit",
    type=int,
    required=False,
    help="Fetch captures for only a limited number of sources, and prefetch that batch to fetch from.",
)
def fetch(
    config: Config,
    prefetch_limit: int | None = None,
) -> None:
    from archive_query_log.captures import fetch_captures

    Capture.init(
        using=config.es.client,
        index=config.es.index_captures,
    )
    fetch_captures(
        config=config,
        prefetch_limit=prefetch_limit,
    )


@captures.group("import")
def import_() -> None:
    pass


def _default_data_dir() -> Path | None:
    ceph_dir = Path("/mnt/ceph/storage/")
    return (
        ceph_dir / "data-in-progress/data-research/web-search/archive-query-log/focused/"
        if ceph_dir.is_mount() and ceph_dir.exists()
        else None
    )


@import_.command(help="Import captures from the AQL-22 dataset.")
@argument("data_dir_path",
          type=PathType(path_type=Path, exists=True, file_okay=False,
                        dir_okay=True, readable=True, writable=False,
                        resolve_path=True, allow_dash=False),
          metavar="DATA_DIR", required=True, default=_default_data_dir)
@option("--check-memento/--no-check-memento", default=True)
@option("--search-provider", type=str, envvar="SEARCH_PROVIDER")
@option("--search-provider-index", type=int,
        envvar="SEARCH_PROVIDER_INDEX")
@pass_config
def aql_22(
        config: Config,
        data_dir_path: Path,
        check_memento: bool,
        search_provider: str | None,
        search_provider_index: int | None,
) -> None:
    from archive_query_log.imports.aql22 import import_captures
    Capture.init(using=config.es.client, index=config.es.index_captures)
    import_captures(
        config=config,
        data_dir_path=data_dir_path,
        check_memento=check_memento,
        search_provider=search_provider,
        search_provider_index=search_provider_index,
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_github_topic_deprecated_org_bandwidth_fork_true_1634361552():
    verify_serp_parsing(
        "https://web.archive.org/web/20211016051912id_/https://github.com/search?q=topic%3Adeprecated+org%3ABandwidth+fork%3Atrue",
        "github",
    )


def test_parse_query_github_topic_docker_stack_org_issuu_1549097482():
    verify_serp_parsing(
        "https://web.archive.org/web/20190202085122id_/https://github.com/search?q=topic%3Adocker-stack+org%3Aissuu&type=Users",
        "github",
    )


def test_parse_query_github_topic_web_components_org_github_fork_true_1650012538():
    verify_serp_parsing(
        "https://web.archive.org/web/20220415084858id_/https://github.com/search?q=topic%3Aweb-components+org%3Agithub+fork%3Atrue",
        "github",
    )


def test_parse_query_github_topic_id_org_codercom_1553274306():
    verify_serp_parsing(
        "https://web.archive.org/web/20190322170506id_/https://github.com/search?q=topic%3Aid+org%3Acodercom&type=Repositories",
        "github",
    )


def test_parse_query_github_mastodon_1657854340():
    verify_serp_parsing(
        "https://web.archive.org/web/20220715030540id_/https://github.com/search?q=mastodon&type=registrypackages",
        "github",
    )


def test_parse_query_github_licensing_gethhwid_1662581846():
    verify_serp_parsing(
        "https://web.archive.org/web/20220907201726id_/https://github.com/search?q=licensing_gethhwid&type=repositories",
        "github",
    )


def test_parse_query_github_subrock_1653099784():
    verify_serp_parsing(
        "https://web.archive.org/web/20220521022304id_/https://github.com/search?q=subrock",
        "github",
    )


def test_parse_query_github_jie_shi_de_jing_mi_gong_ye_you_xian_gong_si_1597737683():
    verify_serp_parsing(
        "https://web.archive.org/web/20200818080123id_/https://github.com/search?q=%E6%9D%B0%E5%A3%AB%E5%BE%B7%E7%B2%BE%E5%AF%86%E5%B7%A5%E4%B8%9A%E6%9C%89%E9%99%90%E5%85%AC%E5%8F%B8",
        "github",
    )


def test_parse_query_github_org_zettlr_1591675518():
    verify_serp_parsing(
        "https://web.archive.org/web/20200609040518id_/https://github.com/search?q=org%3AZettlr+",
        "github",
    )


def test_parse_query_github_topic_bootswatch_org_sslcom_1552777180():
    verify_serp_parsing(
        "https://web.archive.org/web/20190316225940id_/https://github.com/search?q=topic%3Abootswatch+org%3ASSLcom&type=Wikis",
        "github",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_ask_peter_krogh_photographer_1184320758():
    verify_serp_parsing(
        "https://web.archive.org/web/20070713095918id_/http://www.ask.com:80/web?q=Peter+Krogh+Photographer&o=0&qsrc=19&o=0",
        "ask",
    )


def test_parse_query_ask_preparation_of_offal_1172710460():
    verify_serp_parsing(
        "https://web.archive.org/web/20070301005420id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Preparation+of+Offal",
        "ask",
    )


def test_parse_query_ask_tye_dye_instructions_1446417925():
    verify_serp_parsing(
        "https://web.archive.org/web/20151101224525id_/http://www.ask.com/web?q=Tye+Dye+Instructions&qsrc=466&o=13194&l=sem&qo=relatedSearchNarrow",
        "ask",
    )


def test_parse_query_ask_rhino_records_1173555865():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310194425id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Rhino+Records",
        "ask",
    )


def test_parse_query_ask_apologist_1172369538():
    verify_serp_parsing(
        "https://web.archive.org/web/20070225021218id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Apologist",
        "ask",
    )


def test_parse_query_ask_japanese_airlines_1173622069():
    verify_serp_parsing(
        "https://web.archive.org/web/20070311140749id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Japanese+Airlines",
        "ask",
    )


def test_parse_query_ask_whelped_definition_1173526805():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310114005id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Whelped+Definition",
        "ask",
    )


def test_parse_query_ask_what_s_california_s_state_bird_1175130217():
    verify_serp_parsing(
        "https://web.archive.org/web/20070329010337id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=What%27s+California%27s+State+Bird",
        "ask",
    )


def test_parse_query_ask_radioactivity_in_medicine_1173728275():
    verify_serp_parsing(
        "https://web.archive.org/web/20070312193755id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Radioactivity+in+Medicine",
        "ask",
    )


def test_parse_query_ask_free_grammer_test_for_employment_1443411312():
    verify_serp_parsing(
        "https://web.archive.org/web/20150928033512id_/http://www.ask.com:80/web?qsrc=1&o=15145&l=sem&q=free+grammer+test+for+employment",
        "ask",
    )


def test_parse_query_ask_meaning_of_integral_site_www_elook_org_1174247342():
    verify_serp_parsing(
        "https://web.archive.org/web/20070318194902id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=meaning+of+integral+site:www.elook.org",
        "ask",
    )


def test_parse_query_ask_week_names_1173639143():
    verify_serp_parsing(
        "https://web.archive.org/web/20070311185223id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Week+Names",
        "ask",
    )


def test_parse_query_ask_olympic_rings_1172716332():
    verify_serp_parsing(
        "https://web.archive.org/web/20070301023212id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Olympic+Rings",
        "ask",
    )


def test_parse_query_ask_what_should_i_wear_on_the_first_day_of_school_1442349823():
    verify_serp_parsing(
        "https://web.archive.org/web/20150915204343id_/http://www.ask.com/web?qsrc=1&o=ffx&l=dir&q=What+should+I+wear+on+the+first+day+of+school%3F+",
        "ask",
    )


def test_parse_query_ask_genus_falco_1172718376():
    verify_serp_parsing(
        "https://web.archive.org/web/20070301030616id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Genus+Falco",
        "ask",
    )


def test_parse_query_ask_what_words_rhyme_with_afraid_1173540494():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310152814id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=What+Words+Rhyme+with+Afraid%3F",
        "ask",
    )


def test_parse_query_ask_facula_site_www_wordreference_com_1173391757():
    verify_serp_parsing(
        "https://web.archive.org/web/20070308220917id_/http://www.ask.com:80/web?o=8001&qsrc=0&l=dir&q=facula+site:www.wordreference.com",
        "ask",
    )


def test_parse_query_ask_lopsided_site_www_usatoday_com_1173623642():
    verify_serp_parsing(
        "https://web.archive.org/web/20070311143402id_/http://www.ask.com:80/web?o=8001&qsrc=0&l=dir&q=lopsided+site:www.usatoday.com",
        "ask",
    )


def test_parse_query_ask_perfect_first_date_1180217842():
    verify_serp_parsing(
        "https://web.archive.org/web/20070526221722id_/http://www.ask.com:80/web?o=1159&qsrc=6&l=zr&q=Perfect+First+Date",
        "ask",
    )


def test_parse_query_ask_tie_dye_patterns_1446417858():
    verify_serp_parsing(
        "https://web.archive.org/web/20151101224418id_/http://www.ask.com/web?q=Tie+Dye+Patterns&qsrc=466&o=13194&l=sem&qo=relatedSearchNarrow",
        "ask",
    )


def test_parse_query_ask_cure_for_boredom_1174210645():
    verify_serp_parsing(
        "https://web.archive.org/web/20070318093725id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Cure+for+Boredom",
        "ask",
    )


def test_parse_query_ask_saiy_phiohom_1525375661():
    verify_serp_parsing(
        "https://web.archive.org/web/20180503192741id_/https://www.ask.com/web?q=Saiy+phiohom&x=49&y=0&qsrc=2300&o=1102&l=dir//",
        "ask",
    )


def test_parse_query_ask_i_ve_site_news_bbc_co_uk_1172693923():
    verify_serp_parsing(
        "https://web.archive.org/web/20070228201843id_/http://www.ask.com:80/web?o=8001&qsrc=0&l=dir&q=i've+site:news.bbc.co.uk",
        "ask",
    )


def test_parse_query_ask_dog_ear_design_1172708874():
    verify_serp_parsing(
        "https://web.archive.org/web/20070301002754id_/http://www.ask.com:80/web?o=8002&qsrc=6&l=dir&q=Dog+Ear+Design",
        "ask",
    )


def test_parse_query_ask_big_tex_trailer_for_sale_1649774735():
    verify_serp_parsing(
        "https://web.archive.org/web/20220412144535id_/https://www.ask.com/web?q=Big+Tex+Trailer+for+Sale&qo=relatedSearchNarrow&o=1526&l=dis",
        "ask",
    )


def test_parse_query_ask_who_framed_roger_rabbit_1473167643():
    verify_serp_parsing(
        "https://web.archive.org/web/20160906131403id_/http://www.ask.com/web?q=Who+framed+roger+rabbit&qsrc=0&o=0&l=dir%2C+gset&qo=homepageSearchBox",
        "ask",
    )


def test_parse_query_ask_faither_site_www_accringtonweb_com_1173391901():
    verify_serp_parsing(
        "https://web.archive.org/web/20070308221141id_/http://www.ask.com:80/web?o=8001&qsrc=0&l=dir&q=faither++site:www.accringtonweb.com",
        "ask",
    )


def test_parse_query_ask_means_mutable_site_www_geek_interview_comrview_com_1173713625():
    verify_serp_parsing(
        "https://web.archive.org/web/20070312153345id_/http://www.ask.com:80/web?o=8001&qsrc=19&l=dir&q=means+mutable+site%3Awww.geek+interview.comrview.com",
        "ask",
    )


def test_parse_query_ask_date_guy_fawkes_death_1173527460():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310115100id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Date+Guy+Fawkes+Death",
        "ask",
    )


def test_parse_query_ask_interior_design_classes_online_1651218822():
    verify_serp_parsing(
        "https://web.archive.org/web/20220429075342id_/https://www.ask.com/web?q=Interior+Design+Classes+Online&qo=relatedSearchNarrow&o=21029&l=dis",
        "ask",
    )


def test_parse_query_ask_cell_biology_1173639673():
    verify_serp_parsing(
        "https://web.archive.org/web/20070311190113id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Cell+Biology",
        "ask",
    )


def test_parse_query_ask_www_ksu_edu_sa_1182367717():
    verify_serp_parsing(
        "https://web.archive.org/web/20070620192837id_/http://www.ask.com:80/web?o=1215&l=dis&q=www.ksu.edu.sa",
        "ask",
    )


def test_parse_query_ask_names_of_reggae_artists_1173527903():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310115823id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Names+of+Reggae+Artists",
        "ask",
    )


def test_parse_query_ask_vitamins_good_after_expiration_date_expires_site_www_cfs_purdue_edu_1173531858():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310130418id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=vitamins+good+after+expiration+date+expires+site:www.cfs.purdue.edu",
        "ask",
    )


def test_parse_query_ask_fachkunde_metall_fur_ausbildung_in_maschinenbauberufen_1461053109():
    verify_serp_parsing(
        "https://web.archive.org/web/20160419080509id_/http://www.ask.com:80/web?l=dir&qsrc=0&o=ffx&q=Fachkunde+Metall+f%C3%BCr+Ausbildung+in+Maschinenbauberufen",
        "ask",
    )


def test_parse_query_ask_weather_report_for_22903_1489829020():
    verify_serp_parsing(
        "https://web.archive.org/web/20170318092340id_/http://www.ask.com/web?q=weather+report+for+22903&qsrc=257&o=10285&l=iac",
        "ask",
    )


def test_parse_query_ask_old_fashioned_labels_1172720557():
    verify_serp_parsing(
        "https://web.archive.org/web/20070301034237id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Old+Fashioned+Labels",
        "ask",
    )


def test_parse_query_ask_scorch_removal_1172719147():
    verify_serp_parsing(
        "https://web.archive.org/web/20070301031907id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Scorch+Removal",
        "ask",
    )


def test_parse_query_ask_means_mutable_site_www_geekinterview_com_1173392177():
    verify_serp_parsing(
        "https://web.archive.org/web/20070308221617id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=means+mutable+site:www.geekinterview.com",
        "ask",
    )


def test_parse_query_ask_batman_movie_1467834503():
    verify_serp_parsing(
        "https://web.archive.org/web/20160706194823id_/http://www.ask.com/web?q=batman+movie&qsrc=0&o=0&l=dir%2C+tr_uid&qo=homepageSearchBox",
        "ask",
    )


def test_parse_query_ask_if_you_break_a_fluorescent_light_bulb_1618021531():
    verify_serp_parsing(
        "https://web.archive.org/web/20210410022531id_/https://www.ask.com/web?q=if+you+break+a+fluorescent+light+bulb&qsrc=19&o=101881&l=dis",
        "ask",
    )


def test_parse_query_ask_five_phases_of_mitosis_1173539024():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310150344id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Five+Phases+of+Mitosis",
        "ask",
    )


def test_parse_query_ask_are_alkali_metals_lighter_than_water_1619387309():
    verify_serp_parsing(
        "https://web.archive.org/web/20210425214829id_/https://www.ask.com/web?q=Are+alkali+metals+lighter+than+water%3F&search=search&qsrc=178&o=ffx&l=dir",
        "ask",
    )


def test_parse_query_ask_failures_of_president_lincoln_1611567628():
    verify_serp_parsing(
        "https://web.archive.org/web/20210125094028id_/https://www.ask.com/web?q=Failures+of+President+Lincoln&qsrc=2384&o=tb&l=zn",
        "ask",
    )


def test_parse_query_ask_smirnoff_label_1173377085():
    verify_serp_parsing(
        "https://web.archive.org/web/20070308180445id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Smirnoff+Label",
        "ask",
    )


def test_parse_query_ask_boxer_ear_crop_1173516809():
    verify_serp_parsing(
        "https://web.archive.org/web/20070310085329id_/http://www.ask.com:80/web?o=8002&qsrc=6&l=dir&q=Boxer+Ear+Crop",
        "ask",
    )


def test_parse_query_ask_diaframma_contraccezione_lang_it_1175070504():
    verify_serp_parsing(
        "https://web.archive.org/web/20070328082824id_/http://www.ask.com:80/web?tt=1&q=diaframma+contraccezione+lang:IT&qsrc=23&nw=False&u=10",
        "ask",
    )


def test_parse_query_ask_caring_1173637789():
    verify_serp_parsing(
        "https://web.archive.org/web/20070311182949id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Caring",
        "ask",
    )


def test_parse_query_ask_http_my_notes_on_blogspot_com_1401243816():
    verify_serp_parsing(
        "https://web.archive.org/web/20140528022336id_/http://www.ask.com/web?qsrc=2&o=ffx&l=dir&q=%22http%3A%2F%2Fmy-notes-on.blogspot.com%2F%22",
        "ask",
    )


def test_parse_query_ask_afraid_of_women_1172618037():
    verify_serp_parsing(
        "https://web.archive.org/web/20070227231357id_/http://www.ask.com:80/web?o=8001&qsrc=6&l=dir&q=Afraid+of+Women",
        "ask",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_brave_chomikuj_1656776694():
    verify_serp_parsing(
        "https://web.archive.org/web/20220702154454id_/https://search.brave.com/search?q=chomikuj&spellcheck=0",
        "brave",
    )


def test_parse_query_brave_kymatica_1652976922():
    verify_serp_parsing(
        "https://web.archive.org/web/20220519161522id_/https://search.brave.com/search?q=Kymatica",
        "brave",
    )


def test_parse_query_brave_spigot_1660718390():
    verify_serp_parsing(
        "https://web.archive.org/web/20220817063950id_/https://search.brave.com/search?q=spigot",
        "brave",
    )


def test_parse_query_brave_import_esbuild_from_esbuild_syntaxerror_unexpected_identifier_1652662700():
    verify_serp_parsing(
        "https://web.archive.org/web/20220516005820id_/https://search.brave.com/search?q=import+esbuild+from+%22esbuild%22%3B++++++++%5E%5E%5E%5E%5E%5E%5E++SyntaxError%3A+Unexpected+identifier",
        "brave",
    )


def test_parse_query_brave_pizza_box_1662098145():
    verify_serp_parsing(
        "https://web.archive.org/web/20220902055545id_/https://search.brave.com/search?q=pizza%20box",
        "brave",
    )


def test_parse_query_brave_nvfbc_1651575025():
    verify_serp_parsing(
        "https://web.archive.org/web/20220503105025id_/https://search.brave.com/search?q=NvFBC&source=desktop",
        "brave",
    )


def test_parse_query_brave_jinniebella_1666397597():
    verify_serp_parsing(
        "https://web.archive.org/web/20221022001317id_/https://search.brave.com/search?q=jinniebella+&source=web",
        "brave",
    )


def test_parse_query_brave_67_degrees_c_static_atmosphere_1665006071():
    verify_serp_parsing(
        "https://web.archive.org/web/20221005214111id_/https://search.brave.com/search?q=67%20degrees%20C%20static%20atmosphere",
        "brave",
    )


def test_parse_query_brave_new_mac_os_release_1664483322():
    verify_serp_parsing(
        "https://web.archive.org/web/20220929202842id_/https://search.brave.com/search?q=new%20mac%20os%20release",
        "brave",
    )


def test_parse_query_brave_bivoks_trebehandling_1654325149():
    verify_serp_parsing(
        "https://web.archive.org/web/20220604064549id_/https://search.brave.com/search?q=bivoks+trebehandling",
        "brave",
    )


def test_parse_query_brave_google_play_store_1649985658():
    verify_serp_parsing(
        "https://web.archive.org/web/20220415012058id_/https://search.brave.com/search?q=google+play+store&source=web",
        "brave",
    )


def test_parse_query_brave_pekara_fontana_niksic_1649235416():
    verify_serp_parsing(
        "https://web.archive.org/web/20220406085656id_/https://search.brave.com/search?q=pekara+fontana+nik%C5%A1i%C4%87",
        "brave",
    )


def test_parse_query_brave_avnon_dan_1998_martin_buber_the_hidden_dialogue_lanham_boulder_new_york_oxford_rowman_littlefield_1649576842():
    verify_serp_parsing(
        "https://web.archive.org/web/20220410074722id_/https://search.brave.com/search?q=Avnon%2C+Dan%2C+1998%2C+Martin+Buber.+The+Hidden+Dialogue%2C+Lanham%2C+Boulder%2C+New+York%2C+Oxford%3A+Rowman+%26+Littlefield.",
        "brave",
    )


def test_parse_query_brave_metaverse_1650224042():
    verify_serp_parsing(
        "https://web.archive.org/web/20220417193402id_/https://search.brave.com/search?q=%EF%BF%BDmetaverse",
        "brave",
    )


def test_parse_query_brave_h_m_home_1664916031():
    verify_serp_parsing(
        "https://web.archive.org/web/20221004204031id_/https://search.brave.com/search?q=h%26m%20home&spellcheck=0",
        "brave",
    )


def test_parse_query_brave_utilitx_1659649168():
    verify_serp_parsing(
        "https://web.archive.org/web/20220804213928id_/https://search.brave.com/search?q=Utilitx",
        "brave",
    )


def test_parse_query_brave_how_tot_ell_if_tail_recurssive_1658814198():
    verify_serp_parsing(
        "https://web.archive.org/web/20220726054318id_/https://search.brave.com/search?q=how+tot+ell+if+tail+recurssive",
        "brave",
    )


def test_parse_query_brave_random_number_python_1649475875():
    verify_serp_parsing(
        "https://web.archive.org/web/20220409034435id_/https://search.brave.com/search?q=random+number+python&source=desktop",
        "brave",
    )


def test_parse_query_brave_dark_brown_mushrooms_that_melt_in_to_the_ground_1657827723():
    verify_serp_parsing(
        "https://web.archive.org/web/20220714194203id_/https://search.brave.com/search?q=dark+brown+mushrooms+that+melt+in+to+the+ground",
        "brave",
    )


def test_parse_query_brave_a2z_chip_1632000373():
    verify_serp_parsing(
        "https://web.archive.org/web/20210918212613id_/https://search.brave.com/search?q=A2Z+chip",
        "brave",
    )


def test_parse_query_brave_video_1649793956():
    verify_serp_parsing(
        "https://web.archive.org/web/20220412200556id_/https://search.brave.com/search?q=Video",
        "brave",
    )


def test_parse_query_brave_tina_daunt_1652776164():
    verify_serp_parsing(
        "https://web.archive.org/web/20220517082924id_/https://search.brave.com/search?q=tina+daunt",
        "brave",
    )


def test_parse_query_brave_ape_1648509867():
    verify_serp_parsing(
        "https://web.archive.org/web/20220328232427id_/https://search.brave.com/search?q=Ape",
        "brave",
    )


def test_parse_query_brave_nair_1666404375():
    verify_serp_parsing(
        "https://web.archive.org/web/20221022020615id_/https://search.brave.com/search?q=nair",
        "brave",
    )


def test_parse_query_brave_virtualbox_extensions_1667278528():
    verify_serp_parsing(
        "https://web.archive.org/web/20221101045528id_/https://search.brave.com/search?q=virtualbox%20extensions",
        "brave",
    )


def test_parse_query_brave_best_notetaking_app_1654643212():
    verify_serp_parsing(
        "https://web.archive.org/web/20220607230652id_/https://search.brave.com/search?q=best+notetaking+app",
        "brave",
    )


def test_parse_query_brave_logitech_1661137612():
    verify_serp_parsing(
        "https://web.archive.org/web/20220822030652id_/https://search.brave.com/search?q=logitech&source=web",
        "brave",
    )


def test_parse_query_brave_advanced_search_query_1667410105():
    verify_serp_parsing(
        "https://web.archive.org/web/20221102172825id_/https://search.brave.com/search?q=advanced-search-query",
        "brave",
    )


def test_parse_query_brave_froggy_pool_1651139840():
    verify_serp_parsing(
        "https://web.archive.org/web/20220428095720id_/https://search.brave.com/search?q=froggy+pool&source=desktop",
        "brave",
    )


def test_parse_query_brave_new_fold_4_1657628120():
    verify_serp_parsing(
        "https://web.archive.org/web/20220712121520id_/https://search.brave.com/search?q=new+fold+4&source=web",
        "brave",
    )


def test_parse_query_brave_maintenance_1661082520():
    verify_serp_parsing(
        "https://web.archive.org/web/20220821114840id_/https://search.brave.com/search?q=maintenance",
        "brave",
    )


def test_parse_query_brave_product_life_cycle_definition_1664819445():
    verify_serp_parsing(
        "https://web.archive.org/web/20221003175045id_/https://search.brave.com/search?q=product%20life%20cycle%20definition",
        "brave",
    )


def test_parse_query_brave_indiana_vaccine_passport_ban_1629877535():
    verify_serp_parsing(
        "https://web.archive.org/web/20210825074535id_/https://search.brave.com/search?q=Indiana+vaccine+passport+ban",
        "brave",
    )


def test_parse_query_brave_terraform_github_container_1666470147():
    verify_serp_parsing(
        "https://web.archive.org/web/20221022202227id_/https://search.brave.com/search?q=terraform%20github%20container",
        "brave",
    )


def test_parse_query_brave_safemoon_1653585598():
    verify_serp_parsing(
        "https://web.archive.org/web/20220526171958id_/https://search.brave.com/search?q=safemoon",
        "brave",
    )


def test_parse_query_brave_usb_3_1_type_c_dp_alt_mode_1667002333():
    verify_serp_parsing(
        "https://web.archive.org/web/20221029001213id_/https://search.brave.com/search?q=USB+3.1+Type+C+DP+ALT-MODE&source=desktop",
        "brave",
    )


def test_parse_query_brave_copyq_1665902945():
    verify_serp_parsing(
        "https://web.archive.org/web/20221016064905id_/https://search.brave.com/search?q=CopyQ",
        "brave",
    )


def test_parse_query_brave_rent_1647905724():
    verify_serp_parsing(
        "https://web.archive.org/web/20220321233524id_/https://search.brave.com/search?q=Rent",
        "brave",
    )


def test_parse_query_brave_openoffice_1667290210():
    verify_serp_parsing(
        "https://web.archive.org/web/20221101081010id_/https://search.brave.com/search?q=OpenOffice",
        "brave",
    )


def test_parse_query_brave_apple_t3_chip_1651405075():
    verify_serp_parsing(
        "https://web.archive.org/web/20220501113755id_/https://search.brave.com/search?q=apple+t3+chip",
        "brave",
    )


def test_parse_query_brave_how_to_archive_mac_notes_1658260525():
    verify_serp_parsing(
        "https://web.archive.org/web/20220719195525id_/https://search.brave.com/search?q=how+to+archive+mac+notes&source=desktop",
        "brave",
    )


def test_parse_query_brave_amoled_1646168352():
    verify_serp_parsing(
        "https://web.archive.org/web/20220301205912id_/https://search.brave.com/search?q=AMOLED",
        "brave",
    )


def test_parse_query_brave_testimonies_project_1653158644():
    verify_serp_parsing(
        "https://web.archive.org/web/20220521184404id_/https://search.brave.com/search?q=testimonies+project&source=desktop",
        "brave",
    )


def test_parse_query_brave_fastest_otto_guide_1649849902():
    verify_serp_parsing(
        "https://web.archive.org/web/20220413113822id_/https://search.brave.com/search?q=fastest+otto+guide&source=desktop",
        "brave",
    )


def test_parse_query_brave_greentext_1664484724():
    verify_serp_parsing(
        "https://web.archive.org/web/20220929205204id_/https://search.brave.com/search?q=greentext",
        "brave",
    )


def test_parse_query_brave_house_building_cost_estimate_1664320455():
    verify_serp_parsing(
        "https://web.archive.org/web/20220927231415id_/https://search.brave.com/search?q=house%20building%20cost%20estimate",
        "brave",
    )


def test_parse_query_brave_cubecraft_skyblock_pig_farm_1664124008():
    verify_serp_parsing(
        "https://web.archive.org/web/20220925164008id_/https://search.brave.com/search?q=CubeCraft+Skyblock+pig+farm&source=desktop",
        "brave",
    )


def test_parse_query_brave_masi_uno_riser_chain_cover_1658006343():
    verify_serp_parsing(
        "https://web.archive.org/web/20220716211903id_/https://search.brave.com/search?q=masi+uno+riser+chain+cover",
        "brave",
    )


def test_parse_query_brave_unremovable_1650484694():
    verify_serp_parsing(
        "https://web.archive.org/web/20220420195814id_/https://search.brave.com/search?q=unremovable",
        "brave",
    )


def test_parse_query_brave_troll_1649987736():
    verify_serp_parsing(
        "https://web.archive.org/web/20220415015536id_/https://search.brave.com/search?q=troll",
        "brave",
    )



from re import Pattern
from warnings import warn


def clean_text(
        text: str,
        remove_pattern: Pattern | None,
        space_pattern: Pattern | None,
) -> str | None:
    if remove_pattern is not None:
        text = remove_pattern.sub("", text)
    if space_pattern is not None:
        text = space_pattern.sub(" ", text)
    text = text.strip()
    text = " ".join(text.split())
    if text == "":
        return None
    return text


def clean_int(
        text: str,
        remove_pattern: Pattern | None,
) -> int | None:
    if remove_pattern is not None:
        text = remove_pattern.sub("", text)
    text = text.strip()
    try:
        parsed = int(text)
    except ValueError:
        warn(RuntimeWarning(f"Could not parse int: {text}"))
        return None
    return parsed



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_vk_technology_1383988940():
    verify_serp_parsing(
        "https://web.archive.org/web/20131109092220id_/http://vk.com/search?c%5Bq%5D=technology&c%5Bsection%5D=people&offset=0",
        "vk",
    )


def test_parse_query_vk_000_space_marine_1363614912():
    verify_serp_parsing(
        "https://web.archive.org/web/20130318135512id_/http://vk.com/search?c%5Bq%5D=000%3A%20Space%20Marine&c%5Bsection%5D=people&offset=10640",
        "vk",
    )


def test_parse_query_vk_gta_1389323282():
    verify_serp_parsing(
        "https://web.archive.org/web/20140110030802id_/http://vk.com/search?c%5Bq%5D=GTA&c%5Bsection%5D=people&offset=20",
        "vk",
    )


def test_parse_query_vk_radioelektronnye_sistemy_1361402773():
    verify_serp_parsing(
        "https://web.archive.org/web/20130220232613id_/http://vk.com/search?c[name]=0&c[section]=people&c[q]=%D1%80%D0%B0%D0%B4%D0%B8%D0%BE%D1%8D%D0%BB%D0%B5%D0%BA%D1%82%D1%80%D0%BE%D0%BD%D0%BD%D1%8B%D0%B5%20%D1%81%D0%B8%D1%81%D1%82%D0%B5%D0%BC%D1%8B",
        "vk",
    )


def test_parse_query_vk_rabotaiu_na_sebia_1361533546():
    verify_serp_parsing(
        "https://web.archive.org/web/20130222114546id_/http://vk.com/search?c%5Bq%5D=%D1%80%D0%B0%D0%B1%D0%BE%D1%82%D0%B0%D1%8E%20%D0%BD%D0%B0%20%D1%81%D0%B5%D0%B1%D1%8F&c%5Bsection%5D=people&offset=20",
        "vk",
    )


def test_parse_query_vk_imperatory_illiuzii_1387125239():
    verify_serp_parsing(
        "https://web.archive.org/web/20131215163359id_/http://vk.com/search?c[name]=0&c[section]=people&c[q]=%D0%98%D0%BC%D0%BF%D0%B5%D1%80%D0%B0%D1%82%D0%BE%D1%80%D1%8B%20%D0%B8%D0%BB%D0%BB%D1%8E%D0%B7%D0%B8%D0%B9",
        "vk",
    )


def test_parse_query_vk_loco_roco_1377325114():
    verify_serp_parsing(
        "https://web.archive.org/web/20130824061834id_/http://vk.com/search?c[name]=0&c[section]=people&c[q]=Loco%20Roco",
        "vk",
    )


def test_parse_query_vk_sportivnye_mototsikly_1387080107():
    verify_serp_parsing(
        "https://web.archive.org/web/20131215040147id_/http://vk.com/search?c[name]=0&c[section]=people&c[q]=%D1%81%D0%BF%D0%BE%D1%80%D1%82%D0%B8%D0%B2%D0%BD%D1%8B%D0%B5%20%D0%BC%D0%BE%D1%82%D0%BE%D1%86%D0%B8%D0%BA%D0%BB%D1%8B",
        "vk",
    )


def test_parse_query_vk_grot_1353834377():
    verify_serp_parsing(
        "https://web.archive.org/web/20121125090617id_/http://vk.com:80/search?c%5Bq%5D=%D0%B3%D1%80%D0%BE%D1%82&c%5Bsection%5D=auto",
        "vk",
    )


def test_parse_query_vk_muzyka_1372304839():
    verify_serp_parsing(
        "https://web.archive.org/web/20130627034719id_/http://vk.com/search?c%5Bq%5D=%D0%9C%D0%A3%D0%97%D0%AB%D0%9A%D0%90&c%5Bsection%5D=people&offset=0",
        "vk",
    )
